import math
import pdb
from typing import List, Tuple

import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    LambdaLR,
    OneCycleLR,
)

from starling.data.distributions import DiagonalGaussianDistribution
from starling.models import vae_components

torch.set_float32_matmul_precision("high")


class LinearKLDScheduler:
    def __init__(self, total_steps: int, max_weight: float = 1e-4):
        """
        Args:
            total_steps (int): The number of steps to reach the max weight.
            max_weight (float): The maximum weight for the KLD term.
        """
        self.total_steps = total_steps
        self.max_weight = max_weight

    def get_weight(self, current_step: int) -> float:
        """Compute the weight for the current step."""
        if current_step >= self.total_steps:
            return self.max_weight
        return (self.max_weight / self.total_steps) * current_step


class VAE(pl.LightningModule):
    def __init__(
        self,
        model_type: str,
        in_channels: int,
        latent_dim: int,
        dimension: int,
        loss_type: str,
        weights_type: str,
        KLD_weight: float,
        lr_scheduler: str,
        set_lr: float,
        norm: str = "instance",
        base: int = 64,
        optimizer: str = "SGD",
        KLD_warmup_fraction: float = 0.1,
    ) -> None:
        """
        The variational autoencoder (VAE) model that is used to learn the latent space of
        protein distance maps. The model is based on the ResNet architecture and uses a
        Gaussian distribution to model the latent space. The model is trained using the
        evidence lower bound (ELBO) loss, which is a combination of the reconstruction
        loss and the Kullback-Leibler divergence loss. The reconstruction loss can be
        either mean squared error or negative log likelihood. The weights for the
        reconstruction loss can be calculated based on the distance between residues in
        the ground truth distance map. The model can be trained using different learning
        rate schedulers and the learning rate can be set manually.

        References
        ----------
        1) Kingma, D. P. & Welling, M. Auto-Encoding Variational Bayes. arXiv [stat.ML] (2013).

        2) Rombach, R., Blattmann, A., Lorenz, D., Esser, P. & Ommer, B.
        High-resolution image synthesis with latent diffusion models. arXiv [cs.CV] (2021).

        Parameters
        ----------
        model_type : str
            What ResNet architecture to use for the encoder and decoder portion of the VAE
        in_channels : int
            Number of input channels in the input data
        latent_dim : int
            The number of channels in the latent space representation of the data
        dimension : int
            The size of the image in the height and width dimensions (i.e., distance maps)
        loss_type : str
            The type of loss to use for the reconstruction loss. Options are "mse" and "nll"
        weights_type : str
            The type of weights to use for the reconstruction loss. Options are "linear",
            "reciprocal", and "equal"
        KLD_weight : float
            The weight to apply to the KLD loss in the ELBO loss function, KLD loss regularizes the latent space
        lr_scheduler : str
            The learning rate scheduler to use for training the model. Options are "CosineAnnealingWarmRestarts",
            "OneCycleLR", and "CosineAnnealingLR"
        set_lr : float
            The learning rate to use for training the model
        norm : str, optional
            The normalization layer to use in the ResNet architecture, by default "instance"
        base : int, optional
            The base (starting) number of channels to use in the ResNet architecture, by default 64
        optimizer: str, optional
            The optimizer to use in the ResNet architecture, by default "SGD"
        """
        super().__init__()

        self.save_hyperparameters()

        # Set up the ResNet Encoder and Decoder combinations
        resnets = {
            "Resnet18": {
                "encoder": vae_components.Resnet18_Encoder,
                "decoder": vae_components.Resnet18_Decoder,
            },
            "Resnet34": {
                "encoder": vae_components.Resnet34_Encoder,
                "decoder": vae_components.Resnet34_Decoder,
            },
        }

        self.optimizer = optimizer

        # Input dimensions
        self.dimension = dimension

        # Loss params
        self.loss_type = loss_type
        self.weights_type = weights_type

        # Learning rate params
        self.config_scheduler = lr_scheduler
        self.set_lr = set_lr

        # KLD loss params
        self.KLD_weight = KLD_weight
        self.KLD_warmup_fraction = KLD_warmup_fraction

        if self.KLD_warmup_fraction is not None:
            # This needs to be initialized here but will be updated in on_train_start
            self.kl_scheduler = LinearKLDScheduler(
                total_steps=1, max_weight=self.KLD_weight
            )
            self.current_step = 0

        # these are used to monitor the training losses for the *EPOCH*
        self.total_train_step_losses = 0
        self.recon_step_losses = 0
        self.KLD_step_losses = 0
        self.num_batches = 0

        self.monitor = "epoch_val_loss"

        # Encoder
        encoder_chanel = in_channels

        self.encoder = resnets[model_type]["encoder"](
            in_channels=encoder_chanel, base=base, norm=norm
        )

        # This is usually 4 in ResNets
        num_stages = 4
        expansion = self.encoder.block_type.expansion
        exponent = num_stages - 1 if expansion == 1 else num_stages + 1

        # Spatial size of the distance map after the final encoding layer
        self.compressed_size = dimension / (2**num_stages)
        final_channels = int(base * 2**exponent)
        self.shape_from_final_encoding_layer = (
            final_channels,
            self.compressed_size,
            self.compressed_size,
        )

        # Latent space construction layer
        self.encoder_to_latent = nn.Sequential(
            nn.Conv2d(
                final_channels, 2 * latent_dim, kernel_size=3, stride=1, padding=1
            ),
            nn.Conv2d(2 * latent_dim, 2 * latent_dim, kernel_size=1, stride=1),
        )

        # Latent space to decoder layer
        self.latent_to_decoder = nn.Sequential(
            nn.Conv2d(latent_dim, latent_dim, kernel_size=1, stride=1),
            nn.Conv2d(latent_dim, final_channels, kernel_size=3, stride=1, padding=1),
        )

        # Decoder
        decoder_channels = in_channels

        self.decoder = resnets[model_type]["decoder"](
            out_channels=decoder_channels,
            dimension=dimension,
            base=base,
            norm=norm,
        )

        # Params to learn for reconstruction loss
        if self.loss_type == "nll":
            self.log_std = nn.Parameter(torch.zeros(dimension, dimension))

    def encode(self, data: torch.Tensor) -> List[Tuple[torch.Tensor, torch.Tensor]]:
        """
        Takes the data and encodes it into the latent space,
        by returning the mean and log variance

        Parameters
        ----------
        data : torch.Tensor
            Data in the shape of (batch, channel, height, width)

        Returns
        -------
        List[Tuple[torch.Tensor, torch.Tensor]]
            Return the mean and log variance of the latent space
        """

        data = self.encoder(data)

        data = self.encoder_to_latent(data)

        moments = DiagonalGaussianDistribution(data)

        return moments

    def decode(self, latents: torch.Tensor) -> torch.Tensor:
        """
        Decodes the latent space back into the original data

        Parameters
        ----------
        latents : torch.Tensor
            latents in the shape of (batch, channel, height, width)

        Returns
        -------
        torch.Tensor
            Returns the reconstructed data
        """

        latents = self.latent_to_decoder(latents)

        latents = self.decoder(latents)
        return latents

    def reparameterize(self, mu: torch.Tensor, logvar: torch.Tensor) -> torch.Tensor:
        """
        Reparametarization trick that allows for the flow of gradients through the
        non-random process. Check out the paper for more details:
        https://arxiv.org/abs/1312.6114

        Parameters
        ----------
        mu : torch.Tensor
            A tensor containing means of the latent space
        logvar : torch.Tensor
            A tensor containg the log variance of the latent space

        Returns
        -------
        torch.Tensor
            Returns the latent encoding
        """
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + eps * std

    def get_weights(self, ground_truth: torch.Tensor, scale: str) -> torch.Tensor:
        """
        A function that calculates weights for the reconstruction loss based on the
        distance between the residues in the ground truth distance map. The weights
        are calculated based on the scale parameter.

        Parameters
        ----------
        ground_truth : torch.Tensor
            Input data or the ground truth distance map
        scale : str
            A string that determines how the weights will be calculated. The options
            are "linear", "reciprocal", and "equal"

        Returns
        -------
        torch.Tensor
            Returns the weights for the reconstruction loss

        Raises
        ------
        ValueError
            If the scale parameter is not one of the three options
        """
        #! not sure linear is correct rn
        if scale == "linear":
            max_distance = ground_truth.max()
            min_distance = ground_truth.min()
            weights = 1 - (ground_truth - min_distance) / (max_distance - min_distance)
            weights = weights / weights.sum()
            return weights
        # Reciprocal of the distance between the two residues is taken as the weight
        elif scale == "reciprocal":
            # Handling division by zero
            nonzero_indices = ground_truth != 0
            weights = torch.zeros_like(ground_truth)
            weights[nonzero_indices] = torch.reciprocal(ground_truth[nonzero_indices])
            weights = weights / weights.sum()
            return weights
        # We assign equal weight to each distance here
        elif scale == "equal":
            weights = torch.ones_like(ground_truth)
            # Set diagonal elements to zero
            weights = weights - torch.diag(torch.diag(weights))
            weights = weights / weights.sum()
            return weights
        else:
            raise ValueError(f"Variable name '{scale}' for get_weights does not exist")

    def gaussian_likelihood(
        self,
        data_reconstructed: torch.Tensor,
        log_std: torch.Tensor,
        data: torch.Tensor,
    ) -> torch.Tensor:
        """
        Calculates the likelihood of input data given latent space (p(x|z))
        under Gaussian assumption. The reconstructured data is treated as the mean
        of the Gaussian distributions and the log_std is a tensor of learned log standard
        deviations.

        Parameters
        ----------
        data_reconstructed : torch.Tensor
            A tensor containing the reconstructed data that will be treated as the mean to
            parameterize the Gaussian distribution
        log_std : torch.Tensor
            Learned the log standard deviations of the Gaussian distribution
        data : torch.Tensor
            The ground truth data that the likelihood will be calculated against

        Returns
        -------
        torch.Tensor
            Returns the likelihood of the input data given the latent space
        """

        # Create the normal distributions
        dist = torch.distributions.Normal(data_reconstructed, torch.exp(log_std))

        # Calculate log probability of seeing image under p(x|z)
        log_pxz = dist.log_prob(data)

        return log_pxz

    def vae_loss(
        self,
        data_reconstructed: torch.Tensor,
        data: torch.Tensor,
        mu: torch.Tensor,
        logvar: torch.Tensor,
    ) -> dict:
        """
        Calculates the loss of the VAE, using the sum between the KLD loss
        of the latent space to N(0, I) and either mean squared error
        between the reconstructed data and the ground truth or
        the negative log likelihood of the input data given the latent space
        under a Gaussian assumption. Additional loss is added to ensure the
        contacts are reconstructed correctly.

        Parameters
        ----------
        data_reconstructed : torch.Tensor
            Reconstructed data; output of the VAE
        data : torch.Tensor
            Ground truth data, input to the VAE
        mu : torch.Tensor
            Means of the normal distributions of the latent space
        logvar : torch.Tensor
            Log variances of the normal distributions of the latent space
        KLD_weight : int, optional
            How much to weight the importance of the regularization term of the
            latent space. Setting this to lower than 1 will lead to less regular
            and interpretable latent space, by default None

        Returns
        -------
        dict
            Returns a dictionary containing the total loss, reconstruction loss, and KLD loss

        Raises
        ------
        ValueError
            If the loss type is not mse or elbo
        """

        # Find out where 0s are in the data
        mask = (data != 0).float()
        # Remove the lower triangle of the mask so that loss is only calculated on the upper triangle of the distance map
        mask = mask - mask.tril()

        # Mean squared error weighted by ground truth distance
        if self.loss_type == "mse":
            recon = F.mse_loss(data_reconstructed, data, reduction="none")

        # Negative log likelihood of the input data given the latent space
        elif self.loss_type == "nll":
            # Get the reconstruction loss and convert it to positive values
            recon = -1 * self.gaussian_likelihood(
                data_reconstructed=data_reconstructed,
                log_std=self.log_std,
                data=data,
            )
        else:
            raise ValueError(
                f"loss type of name '{self.loss_type}' does not exist. Current implementations include 'mse' and 'nll'"
            )

        # Calculate the loss of only part of the distance map and take the mean
        recon = recon * mask
        recon = torch.sum(recon) / torch.sum(mask)

        # For more information of KLD loss check out Appendix B:
        # https://arxiv.org/abs/1312.6114
        KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp(), dim=[1, 2, 3])
        KLD = torch.logsumexp(KLD, dim=0) / mu.size(0)  # Mean over batch

        if self.KLD_warmup_fraction is not None:
            KLD_weight = self.kl_scheduler.get_weight(self.current_step)
            self.current_step += 1
        else:
            KLD_weight = self.KLD_weight

        loss = recon + KLD_weight * KLD

        return {"loss": loss, "recon": recon, "KLD": KLD}

    def forward(
        self,
        data: torch.Tensor,
    ) -> List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
        """
        Forward pass of the VAE

        Parameters
        ----------
        data : torch.Tensor
            Data in the shape of (batch, channel, height, width) to pass through the VAE

        Returns
        -------
        List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]
            Returns the reconstructed data, the mean of the latent space, and the log variance
        """

        moments = self.encode(data)

        latent_encoding = moments.sample()

        data_reconstructed = self.decode(latent_encoding)

        return data_reconstructed, moments

    def training_step(self, batch: dict, batch_idx) -> torch.Tensor:
        """
        Training step of the VAE compatible with Pytorch Lightning

        Parameters
        ----------
        batch : dict
            A batch of data read in using the DataLoader
        batch_idx : _type_
            Batch number the model is on during training

        Returns
        -------
        torch.Tensor
            Total training loss of this batch
        """
        data = batch

        data_reconstructed, moments = self.forward(data=data)

        loss = self.vae_loss(
            data_reconstructed=data_reconstructed,
            data=data,
            mu=moments.mean,
            logvar=moments.logvar,
        )

        self.total_train_step_losses += loss["loss"].item()
        self.recon_step_losses += loss["recon"].item()
        self.KLD_step_losses += loss["KLD"].item()
        self.num_batches += 1

        self.log("train_loss", loss["loss"], prog_bar=True, batch_size=data.size(0))
        self.log("recon_loss", loss["recon"], prog_bar=True, batch_size=data.size(0))
        self.log("KLD_loss", loss["KLD"], prog_bar=False, batch_size=data.size(0))
        self.log(
            "KLD_weight",
            self.kl_scheduler.get_weight(self.current_step)
            if self.KLD_warmup_fraction is not None
            else self.KLD_weight,
            prog_bar=True,
            batch_size=data.size(0),
        )

        return loss["loss"]

    def on_train_epoch_end(self) -> None:
        """
        At the end of the epoch, calculate the mean of the training losses.
        Clear the lists that have been filled with losses during the epoch
        for memory management.
        """
        epoch_mean = self.total_train_step_losses / self.num_batches
        self.log("epoch_train_loss", epoch_mean, prog_bar=True, sync_dist=True)

        recon_mean = self.recon_step_losses / self.num_batches
        self.log("epoch_recon_loss", recon_mean, prog_bar=True, sync_dist=True)

        KLD_mean = self.KLD_step_losses / self.num_batches
        self.log("epoch_KLD_loss", KLD_mean, prog_bar=True, sync_dist=True)

        # Reset the total losses
        self.total_train_step_losses = 0
        self.recon_step_losses = 0
        self.KLD_step_losses = 0
        self.num_batches = 0

    def validation_step(self, batch: torch.Tensor, batch_idx) -> torch.Tensor:
        """
        Validation step of the VAE compatible with Pytorch Lightning. This is
        called after each epoch.

        Parameters
        ----------
        batch : torch.Tensor
            A batch of data read in using the DataLoader
        batch_idx : _type_
            Batch number the model is on during the validation of the model

        Returns
        -------
        torch.Tensor
            Total validation loss of this batch
        """

        data = batch

        data_reconstructed, moments = self.forward(data=data)

        loss = self.vae_loss(
            data_reconstructed=data_reconstructed,
            data=data,
            mu=moments.mean,
            logvar=moments.logvar,
        )

        self.log(
            "epoch_val_loss",
            loss["loss"],
            prog_bar=True,
            sync_dist=True,
            batch_size=data.size(0),
        )

        return loss["loss"]

    def configure_optimizers(self):
        """
        Configure the optimizer and the learning rate scheduler for the model.
        Here I am using NVIDIA suggested settings for learning rate and weight
        decay. For ResNet50 they have seen best performance with CosineAnnealingLR,
        initial learning rate of 0.256 for batch size of 256 and linearly scaling
        it down/up for other batch sizes. The weight decay is set to 1/32768 for all
        parameters except the batch normalization layers. For further information check:
        https://catalog.ngc.nvidia.com/orgs/nvidia/resources/resnet_50_v1_5_for_pytorch

        Returns
        -------
        List
            Returns the optimizer and the learning rate scheduler

        Raises
        ------
        ValueError
            If the scheduler is not implemented
        """

        optimizer_params = [
            {
                "params": [
                    param
                    for name, param in self.named_parameters()
                    if not any(nd in name for nd in ["bn"]) and name != "log_std"
                ],
                "weight_decay": 1 / 32768,  # Include weight decay for other parameters
            },
            {
                "params": [
                    param
                    for name, param in self.named_parameters()
                    if any(nd in name for nd in ["bn"])
                ],
                "weight_decay": 0.0,  # Exclude weight decay for parameters with 'bn' in name
            },
        ]

        # Conditionally add parameter group for log_std based on self.loss_type
        if self.loss_type == "nll":
            optimizer_params.append(
                {
                    "params": [self.log_std],  # Separate parameter group for log_std
                    "weight_decay": 0.0,  # Exclude weight decay for log_std
                }
            )

        if self.optimizer == "SGD":
            optimizer = torch.optim.SGD(
                optimizer_params,
                lr=self.set_lr,
                momentum=0.875,
                nesterov=True,
            )

        elif self.optimizer == "AdamW":
            optimizer = torch.optim.AdamW(
                optimizer_params,
                lr=self.set_lr,
                betas=(0.9, 0.999),
                eps=1e-08,
            )
        else:
            raise NotImplementedError("Optimizer has not been implemented")

        if self.config_scheduler == "CosineAnnealingWarmRestarts":
            lr_scheduler = {
                "scheduler": CosineAnnealingWarmRestarts(
                    optimizer, T_0=5, eta_min=1e-4
                ),
                "monitor": self.monitor,
                "interval": "epoch",
            }

        elif self.config_scheduler == "OneCycleLR":
            lr_scheduler = {
                "scheduler": OneCycleLR(
                    optimizer,
                    max_lr=0.01,
                    total_steps=self.trainer.estimated_stepping_batches,
                ),
                "monitor": self.monitor,
                "interval": "step",
            }
        elif self.config_scheduler == "LinearWarmupCosineAnnealingLR":
            num_epochs = self.trainer.max_epochs
            total_steps = self.trainer.estimated_stepping_batches
            steps_per_epoch = total_steps // num_epochs
            # Warmup for 5% of the total steps
            warmup_steps = steps_per_epoch * int(num_epochs * 0.05)

            def lr_lambda(current_step):
                if current_step < warmup_steps:
                    # Linear warmup phase
                    return current_step / max(1, warmup_steps)
                else:
                    # Cosine annealing phase
                    eta_min = 1e-5
                    remaining_steps = current_step - warmup_steps
                    current_epoch = remaining_steps // steps_per_epoch
                    cosine_factor = 0.5 * (
                        1 + math.cos(math.pi * current_epoch / num_epochs)
                    )
                    return eta_min + (1 - eta_min) * cosine_factor

            lr_scheduler = {
                "scheduler": LambdaLR(optimizer, lr_lambda=lr_lambda),
                "monitor": self.monitor,
                "interval": "step",
            }
        elif self.config_scheduler == "CosineAnnealingLR":
            num_epochs = self.trainer.max_epochs
            lr_scheduler = {
                "scheduler": CosineAnnealingLR(
                    optimizer,
                    T_max=num_epochs,
                    eta_min=1e-4,
                ),
                "monitor": self.monitor,
                "interval": "epoch",
            }
        else:
            raise ValueError(f"{self.config_scheduler} lr_scheduler is not implemented")

        return [optimizer], [lr_scheduler]

    def symmetrize(self, data_reconstructed: torch.Tensor) -> torch.Tensor:
        """
        Symmetrizes the reconstructed data so that the weights can learn other patterns.
        Loss calculated only on the reconstruction faithfulness of the upper triangle
        of the distance map

        Parameters
        ----------
        data_reconstructed : torch.Tensor
            Reconstructed data; output of the decoder

        Returns
        -------
        torch.Tensor
            Symmetric version of the reconstructed data
        """
        # Get the upper triangular part of each tensor in the batch
        upper_triangles = torch.triu(data_reconstructed)

        # Symmetrize each tensor in the batch individually
        symmetrized_arrays = upper_triangles + torch.transpose(upper_triangles, -1, -2)

        # Fill diagonal elements with zeros for each tensor individually
        diag_values = torch.diagonal(symmetrized_arrays, dim1=-2, dim2=-1)
        symmetrized_arrays = symmetrized_arrays - torch.diag_embed(diag_values)

        return symmetrized_arrays

    # def on_fit_start(self):
    #     def init_weights(m):
    #         if isinstance(m, (nn.Linear, nn.Conv2d)):
    #             nn.init.kaiming_normal_(m.weight, nonlinearity="relu")
    #             if m.bias is not None:
    #                 nn.init.constant_(m.bias, 0)

    #     # Initialize weights for encoder and decoder
    #     self.encoder.apply(init_weights)
    #     self.decoder.apply(init_weights)
    #     # Apply initialization to the encoder_to_latent and latent_to_decoder layers (as they are part of the model)
    #     self.encoder_to_latent.apply(init_weights)
    #     self.latent_to_decoder.apply(init_weights)

    def on_train_start(self):
        if self.KLD_warmup_fraction is not None:
            # Access the trainer and total steps after training starts
            total_steps = self.trainer.estimated_stepping_batches
            # Calculate the warm-up steps (fraction of the total steps)
            kl_warmup_steps = int(total_steps * self.KLD_warmup_fraction)

            # Initialize the KLD scheduler with warm-up steps
            self.kl_scheduler = LinearKLDScheduler(
                total_steps=kl_warmup_steps, max_weight=self.KLD_weight
            )



import matplotlib.pyplot as plt

AA_COLOR = {'Y': '#ff9d00',
            'W': '#ff9d00',
            'F': '#ff9d00',
            'A': '#171616',
            'L': '#171616',
            'M': '#171616',
            'I': '#171616',
            'V': '#171616',
            'Q': '#04700d',
            'N': '#04700d',
            'S': '#04700d',
            'T': '#04700d',
            'H': '#04700d',
            'G': '#04700d',
            'E': '#ff0d0d',
            'D': '#ff0d0d',
            'R': '#2900f5',
            'K': '#2900f5',
            'C': '#ffe70d',
            'P': '#cf30b7'}

def plot_matrices(original, computed, difference, filename):

    """Plot the original, computed, and difference matrices using imshow and save to disk."""
    fig, axs = plt.subplots(1, 3, figsize=(18, 6))

    # Plot the original distance matrix
    im0 = axs[0].imshow(original, cmap="viridis")
    axs[0].set_title("Original Distance Matrix")
    axs[0].set_xlabel("Residue Index")
    axs[0].set_ylabel("Residue Index")
    fig.colorbar(im0, ax=axs[0])

    # Plot the computed distance matrix
    im1 = axs[1].imshow(computed, cmap="viridis")
    axs[1].set_title("Computed Distance Matrix (GD)")
    axs[1].set_xlabel("Residue Index")
    axs[1].set_ylabel("Residue Index")
    fig.colorbar(im1, ax=axs[1])

    # Plot the difference matrix
    im2 = axs[2].imshow(difference, cmap="viridis")
    axs[2].set_title("Difference Matrix (GD)")
    axs[2].set_xlabel("Residue Index")
    axs[2].set_ylabel("Residue Index")
    fig.colorbar(im2, ax=axs[2])

    # Save the plot to disk
    plt.savefig(filename)
    plt.close()

import matplotlib.pyplot as plt
import numpy as np
from sparrow.data.amino_acids import VALID_AMINO_ACIDS


def plot_protein_arcs(sequence, 
                      links, 
                      color_dict=AA_COLOR, 
                      contacts=False,
                      norm_lower_thresh = 1,
                      number_freq = 10,
                      filename=None):
    
    """
    Plot a protein sequence with arcs connecting residues.

    Parameters
    ----------
    sequence : str
        The protein sequence to plot.

    links : list
        A list of links between residues. Each link is represented as a list
        with three elements: the start residue index, the end residue index,
        and the contact score. The contact score can either be a contact score 
        (i.e. value between 0 and 1) reporting on the fraction of the ensemble
        in which the residues are in contact. It can also be a normalized distance; 
        typically the ratio of the actual distances normalized by some limiting
        theoretical model distance.

    contacts : bool, optional
        If True, the contact scores are interpreted as contact scores. If False,
        the contact scores are interpreted as distances. Default is False.

    norm_lower_thresh : float, optional
        The threshold for the lower bound of the normalized distance. Default is 1.

    number_freq : int
        The frequency of the number labels on the x-axis. Default is 10.

    Returns
    -------
    None, but if filename is provided, the plot is saved to disk.

    """
    
    # initialize the plot
    fig, ax = plt.subplots(figsize=(len(sequence) / 4, 4))

    # Set up the plot
    ax.set_xlim(-1, len(sequence))
    ax.set_ylim(-1.75, 1.75)  # Adjusted limits for better visibility

    # this code sets every 10th residue to be labeled
    ax.set_xticks(range(len(sequence)))
    xtick_numberlabels = []
    for idx in range(1, len(sequence)+1):
        if idx % number_freq == 0:
            xtick_numberlabels.append(idx)
        else:
            xtick_numberlabels.append('')            
    ax.set_xticklabels(xtick_numberlabels)  # Hide default labels
        
    # get unique links
    links = list(map(list, set(map(tuple, links))))

    # Color and write each residue to the plot
    for i, res in enumerate(sequence):
        ax.text(i, 0, res, fontsize=12, ha='center', va='center',
                color=color_dict.get(res, 'black'), fontweight='bold')
    

    # if there's at least one link to plot...
    if len(links) > 0:

        # if we passed contact socres (summed or averages)
        if contacts:
    
            # get all the link contact scores
            values = np.array(links)[:,2]
    
            # extract min and max
            min_val = min(values)
            max_val = max(values)
        
            # Avoid division by zero in case all values are the same 
            if max_val == min_val:
                return [0.7] * len(values)  # If all values are the same, map them to the midpoint
        
            # Normalize values to [0, 1], then invert and scale to [0.5, 0.9]
            recoded_links_vals = [0.5 + 0.4 * (max_val - v) / (max_val - min_val) for v in values]
    
            # finally, re-assign in the links list
            for i in range(len(links)):
                links[i][2] = recoded_links_vals[i]
            
        
        # Draw arcs
        idx = -1
        for start, end, score in links:
            idx = idx + 1
            if start >= len(sequence) or end >= len(sequence) or start < 0 or end < 0:
                continue  # Ensure valid indices
            
            mid = (start + end) / 2
            height = min(abs(end - start) / 15, 1)  # Reduce arc height further
    
    
            # if contacts all attractive and we alternative sides
            if contacts:
                color = '#310451'  # Attractive interaction
    
                if idx % 2 == 0:
                    y_offset = -0.1  # Bring arc closer to letters
                    arc = np.linspace(np.pi, 2 * np.pi, 100)  # Flip arc to face downward
                else:
                    y_offset = 0.1  # Bring arc closer to letters
                    arc = np.linspace(0, np.pi, 100)  # Standard arc shape
                    
                intensity = 1 - (score - 0.5) / 0.49  # Scale alpha (0.5 -> 1, 0.99 -> 0.1)
                width = 3 - 2.5 * (score - 0.5) / 0.49  # Scale width (0.5 -> 3, 0.99 -> 0.5)
    
    
            else:
                if score < norm_lower_thresh:
                    color = '#310451'  # Attractive interaction
                    y_offset = 0.1  # Bring arc closer to letters
                    arc = np.linspace(0, np.pi, 100)  # Standard arc shape
                    intensity = 1 - (score - 0.7) / 0.49  # Scale alpha (0.7 -> 1, 0.99 -> 0.1)
                    width = 3 - 2.5 * (score - 0.7) / 0.49  # Scale width (0.7 -> 3, 0.99 -> 0.5)
                else:
                    color = '#813C08'  # Repulsive interaction
                    y_offset = -0.1  # Bring arc closer to letters
                    arc = np.linspace(np.pi, 2 * np.pi, 100)  # Flip arc to face downward
                    intensity = 1 - ((1.5 - score) / 0.49)  # Scale alpha (1.5 -> 1, 1.01 -> 0.1)
                    width = 3 - 2.5 * ((1.5 - score) / 0.49)  # Scale width (1.5 -> 3, 1.01 -> 0.5)
                
            intensity = max(0.1, min(1, intensity))  # Clamp alpha values
            width = max(0.5, min(3, width))  # Clamp width values
            
            # Create arc
            x_arc = np.linspace(start, end, 100)
            y_arc = np.sin(arc) * height + y_offset

            # Create arc
            x_arc = np.linspace(start, end, 100)    
            y_arc = np.sin(arc) * height + y_offset
            ax.plot(x_arc, y_arc, color=color, alpha=intensity, linewidth=width)
        
    
    ax.set_yticks([])
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    
    if filename:
        plt.savefig(filename)
    else:
        plt.show()  




import multiprocessing as mp
from argparse import ArgumentParser
from collections import OrderedDict

import h5py
import hdf5plugin
import numpy as np
import pandas as pd
import torch
import torch.nn.functional as F
from finches.forcefields.mpipi import Mpipi_model, harmonic
from tabulate import tabulate
from tqdm import tqdm

from starling.models.vqvae import VQVAE


def int_to_seq(int_seq):
    """
    Convert an integer sequence to a string sequence.
    """
    aa_to_int = {
        "0": 0,
        "A": 1,
        "C": 2,
        "D": 3,
        "E": 4,
        "F": 5,
        "G": 6,
        "H": 7,
        "I": 8,
        "K": 9,
        "L": 10,
        "M": 11,
        "N": 12,
        "P": 13,
        "Q": 14,
        "R": 15,
        "S": 16,
        "T": 17,
        "V": 18,
        "W": 19,
        "Y": 20,
    }
    reversed_dict = {v: k for k, v in aa_to_int.items()}
    seq = ""
    for i in int_seq:
        if i != 0:
            seq += str(reversed_dict[i])
    return seq


def symmetrize(dm):
    """
    Symmetrize a distance map.
    """
    dm = np.array([np.triu(m, k=1) + np.triu(m, k=1).T for m in dm])

    return dm


def finches_potential_energy(data):
    mpipi = Mpipi_model()
    interaction_energy = 0
    dm = data[0]
    seq = data[1]
    bonds = np.diagonal(dm, offset=1)
    harmonic_energy = harmonic(bonds).sum()
    sequence = list(seq)
    for num, residue in enumerate(sequence):
        for next_residue in range(2, len(sequence[num:])):
            residue_interaction = mpipi.compute_full_Mpipi(
                residue,
                sequence[num + next_residue],
                dm[num, num + next_residue],
            )
            interaction_energy += residue_interaction
    return interaction_energy, harmonic_energy, interaction_energy + harmonic_energy


def load_hdf5_compressed(file_path, keys_to_load=None):
    """
    Loads data from an HDF5 file.

    Parameters:
        - file_path (str): Path to the HDF5 file.
        - keys_to_load (list): List of keys to load. If None, loads all keys.
    Returns:
        - dict: Dictionary containing loaded data.
    """
    data_dict = {}
    with h5py.File(file_path, "r") as f:
        keys = keys_to_load if keys_to_load else f.keys()
        for key in keys:
            if key == "dm":
                data_dict[key] = f[key][...]
            else:
                data_dict[key] = f[key][...]
    return data_dict


def reconstruct(model, distance_maps):
    recon_dm, _ = model(distance_maps)
    recon_dm = recon_dm.detach().cpu().numpy().squeeze()

    return recon_dm


def get_errors(recon_dm, dm, mask):
    recon = F.mse_loss(recon_dm, dm, reduction="none")

    recon = recon * mask
    all_mse = recon.sum(axis=(1, 2)) / mask.sum(axis=(1, 2))

    all_mse = np.array([i.item() for i in all_mse])

    recon_bonds = [i.diagonal(offset=1) for i in recon]
    mask_bonds = [i.diagonal(offset=1) for i in mask]

    bonds_mse = np.array(
        [(i.sum() / j.sum()).item() for i, j in zip(recon_bonds, mask_bonds)]
    )

    return all_mse, bonds_mse


def read_input_file(file_path):
    """
    Read the input file and return a list of paths to the HDF5 files.
    """
    paths = []
    with open(file_path, "r") as f:
        for line in f:
            line = line.strip()
            paths.append(line)
    return paths


def prepare_data(data):
    """
    Prepare the data for inference.
    """
    dm = torch.from_numpy(data)
    dm = dm.unsqueeze(1)

    return dm


def main():
    parser = ArgumentParser()
    parser.add_argument("--input", type=str, required=True)
    parser.add_argument("--vae", type=str, required=True)
    parser.add_argument("--device", type=str, default="cuda:0")
    parser.add_argument("--batch", type=int, default=100)
    parser.add_argument("--outfile", type=str, default="summary_stats_vae.csv")
    args = parser.parse_args()

    # Get the number of cores
    num_cores = mp.cpu_count()
    # Create a pool of workers
    pool = mp.Pool(num_cores)

    # Load the VQVAE model
    vae = VQVAE.load_from_checkpoint(args.vae, map_location=args.device)

    # Read the input file
    paths = read_input_file(args.input)

    # Start a dataframe
    results = OrderedDict()

    for path in tqdm(paths):
        sequence_stats = OrderedDict()
        data = load_hdf5_compressed(path, keys_to_load=["dm", "seq"])

        # Get the data
        data["seq"] = int_to_seq(data["seq"])
        ground_truth_dm = prepare_data(data["dm"])

        num_batches = ground_truth_dm.shape[0] // args.batch
        remaining_samples = ground_truth_dm.shape[0] % args.batch

        recon_dm = []

        for batch in range(num_batches):
            recon_dm.append(
                reconstruct(
                    vae,
                    ground_truth_dm[batch * args.batch : (batch + 1) * args.batch].to(
                        args.device
                    ),
                )
            )

        if remaining_samples > 0:
            recon_dm.append(
                reconstruct(
                    vae,
                    ground_truth_dm[
                        (batch + 1) * args.batch : (batch + 1) * args.batch
                        + remaining_samples
                    ].to(args.device),
                )
            )

        recon_dm = np.concatenate(recon_dm, axis=0)

        mask = data["dm"] != 0
        mask = mask ^ np.tril(mask)
        all_mse, bonds_mse = get_errors(
            torch.from_numpy(recon_dm), torch.from_numpy(data["dm"]), mask
        )

        ## Calculate potential energy
        # recon_energy_data = zip(
        #    recon_dm, [data["seq"] for _ in range(recon_dm.shape[0])]
        # )
        # ground_truth_energy_data = zip(
        #    data["dm"], [data["seq"] for _ in range(recon_dm.shape[0])]
        # )

        ## Run the calculations in parallel
        # recon_results = pool.map(finches_potential_energy, recon_energy_data)
        # ground_truth_results = pool.map(
        #    finches_potential_energy, ground_truth_energy_data
        # )

        # recon_interaction_energy, recon_harmonic_energy, _ = zip(*recon_results)
        # ground_truth_interaction_energy, ground_truth_harmonic_energy, _ = zip(
        #    *ground_truth_results
        # )

        sequence_stats["mse"] = round(all_mse.mean(), 4)
        sequence_stats["std_mse"] = round(all_mse.std(), 4)
        sequence_stats["max_mse"] = round(all_mse.max(), 4)

        sequence_stats["bond_mse"] = round(bonds_mse.mean(), 4)
        sequence_stats["std_bond_mse"] = round(bonds_mse.std(), 4)
        sequence_stats["max_bond_mse"] = round(bonds_mse.max(), 4)

        ## Positive difference means the model is generating distance maps that are less stable
        # difference = list(recon_interaction_energy) - np.array(
        #    ground_truth_interaction_energy
        # )

        # sequence_stats["Potential_energy_abe"] = round(difference.mean(), 4)
        # sequence_stats["Max_potential_energy_abe"] = round(difference.max(), 4)

        sequence_stats["Sequence Length"] = mask[0].diagonal(offset=1).sum() + 1

        results[path] = sequence_stats

    results_df = pd.DataFrame(results).T

    # Calculate the mean and max of each column
    mean_values = results_df.mean().round(4)
    max_values = results_df.max()

    results_df.loc["Overall Mean"] = mean_values
    results_df.loc["Overall Max"] = max_values

    results_df.to_csv(args.outfile, index=True)

    summary = results_df.tail(2).reset_index(drop=False)

    formatted_summary = tabulate(
        summary, headers="keys", tablefmt="pipe", floatfmt=".4f", showindex=False
    )

    print(formatted_summary)


if __name__ == "__main__":
    main()



import torch
import torch.nn.functional as F
from einops import rearrange
from torch import nn

from starling.models.normalization import RMSNorm


class LayerNorm(nn.Module):
    r"""LayerNorm that supports two data formats: channels_last (default) or channels_first.
    The ordering of the dimensions in the inputs. channels_last corresponds to inputs with
    shape (batch_size, height, width, channels) while channels_first corresponds to inputs
    with shape (batch_size, channels, height, width).

    Modified from:
    https://github.com/facebookresearch/ConvNeXt/blob/main/models/convnext.py#L119
    It increases memory requirements substantially, unclear if that can be changed
    """

    def __init__(self, normalized_shape, eps=1e-6, data_format="channels_first"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        if self.data_format == "channels_last":
            return F.layer_norm(
                x, self.normalized_shape, self.weight, self.bias, self.eps
            )
        elif self.data_format == "channels_first":
            x = x.permute(0, 2, 3, 1)
            x = F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
            x = x.permute(0, 3, 1, 2)

            return x


class MinPool2d(nn.Module):
    def __init__(self, kernel_size, stride=None, padding=0, dilation=1):
        super(MinPool2d, self).__init__()
        self.kernel_size = kernel_size
        self.stride = stride or kernel_size
        self.padding = padding
        self.dilation = dilation

    def forward(self, x):
        # Perform min pooling using torch.min and torch.nn.functional.max_pool2d
        unpool = nn.functional.max_pool2d(
            -1 * x,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            return_indices=False,
        )
        return -1 * unpool


#! Fix how the activation function is passed in (should be torch.nn.Module not str)
class ResizeConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        norm: torch.nn.Module,
        activation: str,
        padding: int,
        size: int = None,
        scale_factor: int = None,
        mode: str = "nearest",
    ):
        """
        Instead of ConvTranspose2d, this module uses F.interpolate to upsample the input
        tensor and then applies a convolutional layer. This is useful when the output of
        the upsampled tensor is not the same as the input tensor. This module can be used
        in the decoder part of the network. This type of upsampling is known to fix checkerboard
        artifacts that are common in ConvTranspose2d (https://distill.pub/2016/deconv-checkerboard/).

        Parameters
        ----------
        in_channels : int
            The number of input channels
        out_channels : int
            The number of output channels
        kernel_size : int
            The size of the kernel for the convolutional layer
        norm : torch.nn.Module
            The normalization layer to use. Pass in the class of the normalization layer (example nn.InstanceNorm2d)
        activation : str
            The activation function to use. Pass in the class of the activation function (example nn.ReLU)
        padding : int
            The padding to use for the convolutional layer
        size : int, optional
            The spatial size of the output tensor. If None, the scale_factor is used, by default None
        scale_factor : int, optional
            The scale factor to use for the upsampling, by default None
        mode : str, optional
            The mode to use for the upsampling, by default "nearest"
        """
        super().__init__()
        self.size = size
        self.scale_factor = scale_factor
        self.mode = mode

        if norm is not None:
            normalization = (
                norm(out_channels) if norm != nn.GroupNorm else norm(32, out_channels)
            )

        self.conv = nn.Sequential(
            nn.Conv2d(
                in_channels, out_channels, kernel_size, stride=1, padding=padding
            ),
            nn.Identity() if norm is None else normalization,
            nn.Identity() if activation is None else nn.ReLU(inplace=True),
        )

    def forward(self, x):
        x = F.interpolate(
            x, size=self.size, scale_factor=self.scale_factor, mode=self.mode
        )
        x = self.conv(x)
        return x


class ResBlockEncBasic(nn.Module):
    expansion = 1

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int,
        norm: str,
        timestep: int = None,
        kernel_size: int = 3,
    ) -> None:
        """
        A basic residual block. This block is often used in the ResNet architecture,
        specifically in ResNet18 and ResNet34. It consists of two convolutional layers
        with a ReLU activation function in between. The input is added to the output of
        the second convolutional layer. The output is then passed through another ReLU
        activation function. The block can also be conditioned on some class labels or
        other information that is summed to the output of first convolutional output before
        the normalization and activation function.

        Parameters
        ----------
        in_channels : int
            The number of input channels
        out_channels : int
            The number of output channels
        stride : int
            The stride of the first convolutional layer
        norm : str
            The normalization layer to use. Options are "batch", "instance", "layer", "rms" and "group"
        timestep : int, optional
            The dimension of the class labels/timesteps to condition the block on, if None no
            conditioning is done, by default None
        kernel_size : int, optional
            The size of the kernel for convolutional layers, by default 3
        """
        super().__init__()

        kernel_size = 3 if kernel_size is None else kernel_size
        padding = 2 if kernel_size == 5 else (3 if kernel_size == 7 else 1)

        normalization = {
            "batch": nn.BatchNorm2d,
            "instance": nn.InstanceNorm2d,
            "layer": LayerNorm,
            "rms": RMSNorm,
            "group": nn.GroupNorm,
        }

        self.conv1 = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            stride=stride,
            padding=padding,
            kernel_size=kernel_size,
        )
        self.norm1 = (
            normalization[norm](out_channels)
            if norm != "group"
            else normalization[norm](32, out_channels)
        )

        self.activation1 = nn.ReLU(inplace=True)

        if timestep is not None:
            self.time_mlp = nn.Sequential(
                nn.SiLU(inplace=False),
                nn.Linear(timestep, out_channels * 2),
            )

        self.conv2 = nn.Sequential(
            nn.Conv2d(
                in_channels=out_channels,
                out_channels=out_channels,
                stride=1,
                padding=padding,
                kernel_size=kernel_size,
            ),
            normalization[norm](out_channels)
            if norm != "group"
            else normalization[norm](32, out_channels),
        )

        if stride > 1 or in_channels != out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=1,
                    stride=stride,
                    padding=0,
                ),
                normalization[norm](out_channels)
                if norm != "group"
                else normalization[norm](32, out_channels),
            )
        else:
            self.shortcut = nn.Sequential()
        self.activation = nn.ReLU(inplace=True)

    def forward(self, data, timestep=None):
        # Set up the shortcut connection if necessary
        identity = self.shortcut(data)

        # First convolution
        data = self.conv1(data)

        # Add timestep conditioning if provided using FiLM
        if timestep is not None:
            timestep = self.time_mlp(timestep)
            timestep = rearrange(timestep, "b c -> b c 1 1")
            # See the following link for explanation of scale, shift for timestep/class conditioning
            # https://distill.pub/2018/feature-wise-transformations/
            scale, shift = timestep.chunk(2, dim=1)
            data = data * (scale + 1) + shift

        # Add normalization and activation function after timestep conditioning
        data = self.norm1(data)
        data = self.activation1(data)

        # Second convolution
        data = self.conv2(data)

        # Add the input and run it through activation function
        data += identity
        return self.activation(data)


class ResBlockDecBasic(nn.Module):
    contraction = 1

    def __init__(
        self,
        in_channels,
        out_channels,
        stride,
        norm,
        last_layer=None,
        kernel_size=None,
    ) -> None:
        super().__init__()

        kernel_size = 3 if kernel_size is None else kernel_size
        padding = 2 if kernel_size == 5 else (3 if kernel_size == 7 else 1)

        normalization = {
            "batch": nn.BatchNorm2d,
            "instance": nn.InstanceNorm2d,
            "layer": LayerNorm,
            "group": nn.GroupNorm,
        }

        # First convolution which doesn't change the shape of the tensor
        # (b, c, h, w) -> (b, c, h, w) stride = 1
        self.conv1 = nn.Sequential(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=in_channels,
                stride=1,
                padding=padding,
                kernel_size=kernel_size,
            ),
            normalization[norm](in_channels)
            if norm != "group"
            else normalization[norm](32, in_channels),
            nn.ReLU(inplace=True),
        )

        if stride > 1:
            self.conv2 = ResizeConv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                norm=normalization[norm],
                activation=None,
                padding=padding,
                scale_factor=stride,
                mode="nearest",
            )

            self.shortcut = ResizeConv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                norm=normalization[norm],
                activation=None,
                padding=0,
                scale_factor=stride,
                mode="nearest",
            )
        else:
            self.conv2 = nn.Sequential(
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    stride=1,
                    padding=padding,
                    kernel_size=kernel_size,
                ),
                normalization[norm](out_channels)
                if norm != "group"
                else normalization[norm](32, out_channels),
            )
            self.shortcut = nn.Sequential()

        self.activation = nn.ReLU(inplace=True)

    def forward(self, data):
        # Setup the shortcut connection if necessary
        identity = self.shortcut(data)
        # First convolution of the data
        data = self.conv1(data)
        # Second convolution of the data
        data = self.conv2(data)
        # Connect the input data to the output of convolutions
        data += identity
        # Run it through the activation function
        return self.activation(data)


class ResBlockEncBottleneck(nn.Module):
    expansion = 4

    def __init__(
        self,
        in_channels,
        out_channels,
        stride,
        expansion=4,
    ) -> None:
        super().__init__()
        self.expansion = expansion

        self.conv1 = nn.Sequential(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

        self.conv2 = nn.Sequential(
            nn.Conv2d(
                in_channels=out_channels,
                out_channels=out_channels,
                kernel_size=3,
                stride=stride,
                padding=1,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

        self.conv3 = nn.Sequential(
            nn.Conv2d(
                in_channels=out_channels,
                out_channels=int(out_channels * self.expansion),
                kernel_size=1,
            ),
            nn.BatchNorm2d(int(out_channels * self.expansion)),
            nn.ReLU(inplace=True),
        )

        if stride != 1 or in_channels != int(out_channels * self.expansion):
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=int(out_channels * self.expansion),
                    kernel_size=1,
                    stride=stride,
                ),
                nn.BatchNorm2d(int(out_channels * self.expansion)),
            )
        else:
            self.shortcut = nn.Sequential()

        self.activation = nn.ReLU(inplace=True)

    def forward(self, data):
        identity = self.shortcut(data)
        out = self.conv1(data)
        out = self.conv2(out)
        out = self.conv3(out)
        out = out + identity
        return self.activation(out)


class ResBlockDecBottleneck(nn.Module):
    contraction = 4

    def __init__(
        self, in_channels, out_channels, stride, contraction=4, last_layer=False
    ) -> None:
        super().__init__()

        self.contraction = contraction

        self.conv1 = nn.Sequential(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

        if stride != 1:
            self.conv2 = nn.Sequential(
                nn.ConvTranspose2d(
                    in_channels=out_channels,
                    out_channels=out_channels,
                    kernel_size=3,
                    stride=stride,
                    output_padding=1,
                    padding=1,
                ),
                nn.BatchNorm2d(out_channels),
                nn.ReLU(inplace=True),
            )
        else:
            self.conv2 = nn.Sequential(
                nn.Conv2d(
                    in_channels=out_channels,
                    out_channels=out_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                ),
                nn.BatchNorm2d(out_channels),
                nn.ReLU(inplace=True),
            )

        expansion = (
            self.contraction
            if stride == 1 and not last_layer
            else (1 if last_layer else int(self.contraction / 2))
        )
        self.conv3 = nn.Sequential(
            nn.Conv2d(
                in_channels=out_channels,
                out_channels=out_channels * expansion,
                kernel_size=1,
            ),
            nn.BatchNorm2d(out_channels * expansion),
            nn.ReLU(inplace=True),
        )

        if stride != 1 or last_layer:
            expansion = 1 if last_layer else int(self.contraction / 2)
            self.shortcut = nn.Sequential(
                nn.ConvTranspose2d(
                    in_channels=in_channels,
                    out_channels=int(out_channels * expansion),
                    kernel_size=1,
                    stride=stride,
                    output_padding=1 if stride > 1 else 0,
                ),
                nn.BatchNorm2d(int(out_channels * expansion)),
            )
        else:
            self.shortcut = nn.Sequential()

        self.activation = nn.ReLU(inplace=True)

    def forward(self, data):
        identity = self.shortcut(data)
        out = self.conv1(data)
        out = self.conv2(out)
        out = self.conv3(out)
        out = out + identity
        return self.activation(out)


def instance_norm(features, eps=1e-6, **kwargs):
    return nn.InstanceNorm2d(features, affine=True, eps=eps, **kwargs)


def layer_norm(out_channels, starting_dimension, **kwargs):
    denominator = 4 * (out_channels / 64)
    dimension = int(starting_dimension / denominator)
    return nn.LayerNorm([out_channels, dimension, dimension])


class vanilla_Encoder(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride) -> None:
        super().__init__()

        padding = 2 if kernel_size == 5 else (3 if kernel_size == 7 else 1)

        modules = []
        for num, hidden_dim in enumerate(out_channels):
            modules.append(
                nn.Sequential(
                    nn.Conv2d(
                        in_channels,
                        out_channels=hidden_dim,
                        kernel_size=kernel_size,
                        stride=stride,
                        padding=padding,
                    ),
                    nn.BatchNorm2d(hidden_dim),
                    nn.ReLU(inplace=True),
                )
            )
            in_channels = hidden_dim

        modules.append(
            nn.Sequential(
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channels[-1],
                    kernel_size=3,
                    stride=1,
                    padding=0,
                ),
                nn.BatchNorm2d(out_channels[-1]),
                nn.ReLU(inplace=True),
            )
        )

        self.encoder = nn.Sequential(*modules)

    def forward(self, data):
        return self.encoder(data)


class vanilla_Decoder(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride) -> None:
        super().__init__()

        padding = 2 if kernel_size == 5 else (3 if kernel_size == 7 else 1)

        modules = []

        modules.append(
            nn.Sequential(
                nn.ConvTranspose2d(
                    in_channels=out_channels[0],
                    out_channels=out_channels[0],
                    kernel_size=3,
                    stride=1,
                    padding=0,
                ),
                nn.BatchNorm2d(out_channels[0]),
                nn.ReLU(inplace=True),
            )
        )

        num_layers = len(out_channels) - 1
        for num in range(num_layers):
            modules.append(
                nn.Sequential(
                    nn.ConvTranspose2d(
                        out_channels[num],
                        out_channels[num + 1],
                        kernel_size=kernel_size,
                        stride=stride,
                        padding=padding,
                        output_padding=1,
                    ),
                    nn.BatchNorm2d(out_channels[num + 1]),
                    nn.ReLU(inplace=True),
                )
            )

        # Final output layer
        modules.append(
            nn.Sequential(
                nn.Conv2d(
                    in_channels=out_channels[-1],
                    out_channels=out_channels[-1],
                    kernel_size=kernel_size,
                    stride=1,
                    padding=padding,
                ),
                nn.ReLU(inplace=True),
            )
        )

        self.decoder = nn.Sequential(*modules)

    def forward(self, data):
        return self.decoder(data)


class DownsampleBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride):
        super().__init__()

        padding = 2 if kernel_size == 5 else (3 if kernel_size == 7 else 1)

        self.conv = nn.Sequential(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, data):
        return self.conv(data)


class UpsampleBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride):
        super().__init__()

        padding = 2 if kernel_size == 5 else (3 if kernel_size == 7 else 1)

        self.conv_transpose = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                output_padding=1,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, data):
        return self.conv_transpose(data)



import torch
from torch import nn
from torch.nn import functional as F


class RMSNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.g = nn.Parameter(torch.ones(1, dim, 1, 1))

    def forward(self, x):
        return F.normalize(x, dim=1) * self.g * (x.shape[1] ** 0.5)



import numpy as np
import torch


class DiagonalGaussianDistribution(object):
    def __init__(self, parameters: torch.Tensor, deterministic: bool = False):
        """
        Diagonal Gaussian distribution that can be used for sampling from given certain parameters.

        Parameters
        ----------
        parameters : torch.Tensor
            Parameters of the distribution.
        deterministic : bool, optional
            Whether to sample deterministically, by default False
        """

        self.parameters = parameters
        # Split the parameters into mean and logvar
        self.mean, self.logvar = torch.chunk(parameters, 2, dim=1)

        # Clamp the logvar to prevent numerical instability
        #! Maybe switch to (-10, 10) - might help with numerical stability
        #! or self.std = torch.sqrt(torch.nn.functional.softplus(self.logvar))
        self.logvar = torch.clamp(self.logvar, -30.0, 20.0)
        self.deterministic = deterministic

        # Compute the standard deviation and variance
        self.std = torch.exp(0.5 * self.logvar)
        self.var = torch.exp(self.logvar)

        # If deterministic sampling, set the variance to zero
        if self.deterministic:
            self.var = self.std = torch.zeros_like(self.mean).to(
                device=self.parameters.device
            )

    def sample(self) -> torch.Tensor:
        """
        Sample from the parameterized distribution.

        Returns
        -------
        torch.Tensor
            The sampled value.
        """
        x = self.mean + self.std * torch.randn(self.mean.shape).to(
            device=self.parameters.device
        )
        return x

    def kl(self, other=None):
        if self.deterministic:
            return torch.Tensor([0.0])
        else:
            if other is None:
                return 0.5 * torch.sum(
                    torch.pow(self.mean, 2) + self.var - 1.0 - self.logvar,
                    dim=[1, 2, 3],
                )
            else:
                return 0.5 * torch.sum(
                    torch.pow(self.mean - other.mean, 2) / other.var
                    + self.var / other.var
                    - 1.0
                    - self.logvar
                    + other.logvar,
                    dim=[1, 2, 3],
                )

    def nll(self, sample, dims=[1, 2, 3]):
        if self.deterministic:
            return torch.Tensor([0.0])
        logtwopi = np.log(2.0 * np.pi)
        return 0.5 * torch.sum(
            logtwopi + self.logvar + torch.pow(sample - self.mean, 2) / self.var,
            dim=dims,
        )

    def mode(self) -> torch.Tensor:
        """
        Return the mode of the distribution.

        Returns
        -------
        torch.Tensor
            Mode of the distribution.
        """
        return self.mean



import math

import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    LambdaLR,
    OneCycleLR,
)

from starling.models import vae_components
from starling.models.quantize import VectorQuantizer2 as VectorQuantizer


class PrintLayer(nn.Module):
    def __init__(self, layer_name):
        super(PrintLayer, self).__init__()
        self.layer_name = layer_name

    def forward(self, x):
        print(f"Intermediate output of {self.layer_name}: {x.shape}")
        return x


torch.set_float32_matmul_precision("high")


class VQVAE(pl.LightningModule):
    def __init__(
        self,
        model_type: str,
        in_channels: int,
        latent_dim: int,
        dimension: int,
        loss_type: str,
        lr_scheduler: str,
        set_lr: float,
        num_classes: int = 512,
        beta: float = 0.25,
        norm: str = "instance",
        base: int = 64,
        optimizer: str = "SGD",
    ) -> None:
        """
        The variational autoencoder (VAE) model that is used to learn the latent space of
        protein distance maps. The model is based on the ResNet architecture and uses a
        Gaussian distribution to model the latent space. The model is trained using the
        evidence lower bound (ELBO) loss, which is a combination of the reconstruction
        loss and the Kullback-Leibler divergence loss. The reconstruction loss can be
        either mean squared error or negative log likelihood. The weights for the
        reconstruction loss can be calculated based on the distance between residues in
        the ground truth distance map. The model can be trained using different learning
        rate schedulers and the learning rate can be set manually.

        References
        ----------
        1) Kingma, D. P. & Welling, M. Auto-Encoding Variational Bayes. arXiv [stat.ML] (2013).

        2) Rombach, R., Blattmann, A., Lorenz, D., Esser, P. & Ommer, B.
        High-resolution image synthesis with latent diffusion models. arXiv [cs.CV] (2021).

        Parameters
        ----------
        model_type : str
            What ResNet architecture to use for the encoder and decoder portion of the VAE
        in_channels : int
            Number of input channels in the input data
        latent_dim : int
            The number of channels in the latent space representation of the data
        dimension : int
            The size of the image in the height and width dimensions (i.e., distance maps)
        loss_type : str
            The type of loss to use for the reconstruction loss. Options are "mse" and "nll"
        KLD_weight : float
            The weight to apply to the KLD loss in the ELBO loss function, KLD loss regularizes the latent space
        lr_scheduler : str
            The learning rate scheduler to use for training the model. Options are "CosineAnnealingWarmRestarts",
            "OneCycleLR", and "CosineAnnealingLR"
        set_lr : float
            The learning rate to use for training the model
        norm : str, optional
            The normalization layer to use in the ResNet architecture, by default "instance"
        base : int, optional
            The base (starting) number of channels to use in the ResNet architecture, by default 64
        optimizer: str, optional
            The optimizer to use in the ResNet architecture, by default "SGD"
        """
        super().__init__()

        self.save_hyperparameters()

        # Set up the ResNet Encoder and Decoder combinations
        resnets = {
            "Resnet18": {
                "encoder": vae_components.Resnet18_Encoder,
                "decoder": vae_components.Resnet18_Decoder,
            },
            "Resnet34": {
                "encoder": vae_components.Resnet34_Encoder,
                "decoder": vae_components.Resnet34_Decoder,
            },
        }

        self.optimizer = optimizer

        # Input dimensions
        self.dimension = dimension

        # Loss params
        self.loss_type = loss_type

        # Learning rate params
        self.config_scheduler = lr_scheduler
        self.set_lr = set_lr

        # these are used to monitor the training losses for the *EPOCH*
        self.total_train_step_losses = 0
        self.recon_step_losses = 0
        self.quantized_step_losses = 0
        self.num_batches = 0

        self.monitor = "epoch_val_loss"

        # Encoder
        encoder_chanel = in_channels

        self.encoder = resnets[model_type]["encoder"](
            in_channels=encoder_chanel, base=base, norm=norm
        )

        # This is usually 4 in ResNets
        num_stages = 4
        expansion = self.encoder.block_type.expansion
        exponent = num_stages - 1 if expansion == 1 else num_stages + 1

        # Spatial size of the distance map after the final encoding layer
        self.compressed_size = dimension / (2**num_stages)
        final_channels = int(base * 2**exponent)
        self.shape_from_final_encoding_layer = (
            final_channels,
            self.compressed_size,
            self.compressed_size,
        )

        self.quantize = VectorQuantizer(
            n_e=num_classes,
            e_dim=latent_dim,
            beta=beta,
            legacy=False,
        )

        self.quant_conv = nn.Sequential(
            nn.Conv2d(final_channels, latent_dim, kernel_size=3, stride=1, padding=1),
            nn.Conv2d(latent_dim, latent_dim, 1),
        )

        self.post_quant_conv = nn.Sequential(
            nn.Conv2d(latent_dim, latent_dim, kernel_size=1, stride=1),
            torch.nn.Conv2d(
                latent_dim, final_channels, kernel_size=3, stride=1, padding=1
            ),
        )

        # Decoder
        decoder_channels = in_channels

        self.decoder = resnets[model_type]["decoder"](
            out_channels=decoder_channels,
            dimension=dimension,
            base=base,
            norm=norm,
        )

        # Params to learn for reconstruction loss
        if self.loss_type == "nll":
            self.log_std = nn.Parameter(torch.zeros(dimension, dimension))

    def encode(self, data):
        h = self.encoder(data)
        h = self.quant_conv(h)
        quant, emb_loss, info = self.quantize(h)
        return quant, emb_loss, info

    def encode_to_prequant(self, x):
        h = self.encoder(x)
        h = self.quant_conv(h)
        return h

    def decode(self, quant):
        quant = self.post_quant_conv(quant)
        dec = self.decoder(quant)
        return dec

    def decode_code(self, code_b):
        quant_b = self.quantize.embed_code(code_b)
        dec = self.decode(quant_b)
        return dec

    def gaussian_likelihood(
        self,
        data_reconstructed: torch.Tensor,
        log_std: torch.Tensor,
        data: torch.Tensor,
    ) -> torch.Tensor:
        """
        Calculates the likelihood of input data given latent space (p(x|z))
        under Gaussian assumption. The reconstructured data is treated as the mean
        of the Gaussian distributions and the log_std is a tensor of learned log standard
        deviations.

        Parameters
        ----------
        data_reconstructed : torch.Tensor
            A tensor containing the reconstructed data that will be treated as the mean to
            parameterize the Gaussian distribution
        log_std : torch.Tensor
            Learned the log standard deviations of the Gaussian distribution
        data : torch.Tensor
            The ground truth data that the likelihood will be calculated against

        Returns
        -------
        torch.Tensor
            Returns the likelihood of the input data given the latent space
        """

        # Create the normal distributions
        dist = torch.distributions.Normal(data_reconstructed, torch.exp(log_std))

        # Calculate log probability of seeing image under p(x|z)
        log_pxz = dist.log_prob(data)

        return log_pxz

    def vae_loss(
        self,
        data_reconstructed: torch.Tensor,
        data: torch.Tensor,
        quantized_loss: torch.Tensor,
    ) -> dict:
        """
        Calculates the loss of the VAE, using the sum between the KLD loss
        of the latent space to N(0, I) and either mean squared error
        between the reconstructed data and the ground truth or
        the negative log likelihood of the input data given the latent space
        under a Gaussian assumption. Additional loss is added to ensure the
        contacts are reconstructed correctly.

        Parameters
        ----------
        data_reconstructed : torch.Tensor
            Reconstructed data; output of the VAE
        data : torch.Tensor
            Ground truth data, input to the VAE
        mu : torch.Tensor
            Means of the normal distributions of the latent space
        logvar : torch.Tensor
            Log variances of the normal distributions of the latent space
        KLD_weight : int, optional
            How much to weight the importance of the regularization term of the
            latent space. Setting this to lower than 1 will lead to less regular
            and interpretable latent space, by default None

        Returns
        -------
        dict
            Returns a dictionary containing the total loss, reconstruction loss, and KLD loss

        Raises
        ------
        ValueError
            If the loss type is not mse or elbo
        """

        # Find out where 0s are in the data
        mask = (data != 0).float()
        # Remove the lower triangle of the mask so that loss is only calculated on the upper triangle of the distance map
        mask = mask - mask.tril()

        # Mean squared error weighted by ground truth distance
        if self.loss_type == "mse":
            recon = F.mse_loss(data_reconstructed, data, reduction="none")

        # Negative log likelihood of the input data given the latent space
        elif self.loss_type == "nll":
            # Get the reconstruction loss and convert it to positive values
            recon = -1 * self.gaussian_likelihood(
                data_reconstructed=data_reconstructed,
                log_std=self.log_std,
                data=data,
            )
        else:
            raise ValueError(
                f"loss type of name '{self.loss_type}' does not exist. Current implementations include 'mse' and 'nll'"
            )

        # Calculate the loss of only part of the distance map and take the mean
        recon = recon * mask
        recon = torch.sum(recon) / torch.sum(mask)

        loss = recon + quantized_loss

        return {"loss": loss, "recon": recon, "quantized_loss": quantized_loss}

    def forward(self, input, return_pred_indices=False):
        quant, quantized_loss, (_, _, ind) = self.encode(input)
        dec = self.decode(quant)
        if return_pred_indices:
            return dec, quantized_loss, ind
        return dec, quantized_loss

    def training_step(self, batch: dict, batch_idx) -> torch.Tensor:
        """
        Training step of the VAE compatible with Pytorch Lightning

        Parameters
        ----------
        batch : dict
            A batch of data read in using the DataLoader
        batch_idx : _type_
            Batch number the model is on during training

        Returns
        -------
        torch.Tensor
            Total training loss of this batch
        """
        data = batch

        data_reconstructed, quantized_loss = self.forward(input=data)

        loss = self.vae_loss(
            data_reconstructed=data_reconstructed,
            data=data,
            quantized_loss=quantized_loss,
        )

        self.total_train_step_losses += loss["loss"].item()
        self.recon_step_losses += loss["recon"].item()
        self.quantized_step_losses += loss["quantized_loss"].item()
        self.num_batches += 1

        self.log("train_loss", loss["loss"], prog_bar=True, batch_size=data.size(0))
        self.log("recon_loss", loss["recon"], prog_bar=True, batch_size=data.size(0))
        self.log(
            "quantized_loss",
            loss["quantized_loss"],
            prog_bar=True,
            batch_size=data.size(0),
        )

        return loss["loss"]

    def on_train_epoch_end(self) -> None:
        """
        At the end of the epoch, calculate the mean of the training losses.
        Clear the lists that have been filled with losses during the epoch
        for memory management.
        """
        epoch_mean = self.total_train_step_losses / self.num_batches
        self.log("epoch_train_loss", epoch_mean, prog_bar=True, sync_dist=True)

        recon_mean = self.recon_step_losses / self.num_batches
        self.log("epoch_recon_loss", recon_mean, prog_bar=True, sync_dist=True)

        quantized_mean = self.quantized_step_losses / self.num_batches
        self.log("epoch_quantized_loss", quantized_mean, prog_bar=True, sync_dist=True)

        # Reset the total losses
        self.total_train_step_losses = 0
        self.recon_step_losses = 0
        self.quantized_step_losses = 0
        self.num_batches = 0

    def validation_step(self, batch: torch.Tensor, batch_idx) -> torch.Tensor:
        """
        Validation step of the VAE compatible with Pytorch Lightning. This is
        called after each epoch.

        Parameters
        ----------
        batch : torch.Tensor
            A batch of data read in using the DataLoader
        batch_idx : _type_
            Batch number the model is on during the validation of the model

        Returns
        -------
        torch.Tensor
            Total validation loss of this batch
        """

        data = batch

        data_reconstructed, quantized_loss = self.forward(input=data)

        loss = self.vae_loss(
            data_reconstructed=data_reconstructed,
            data=data,
            quantized_loss=quantized_loss,
        )

        # self.log(
        #     "epoch_recon_loss",
        #     loss["recon"],
        #     prog_bar=True,
        #     sync_dist=True,
        #     batch_size=data.size(0),
        # )

        self.log(
            "epoch_val_loss",
            loss["loss"],
            prog_bar=True,
            sync_dist=True,
            batch_size=data.size(0),
        )

        return loss["loss"]

    def configure_optimizers(self):
        """
        Configure the optimizer and the learning rate scheduler for the model.
        Here I am using NVIDIA suggested settings for learning rate and weight
        decay. For ResNet50 they have seen best performance with CosineAnnealingLR,
        initial learning rate of 0.256 for batch size of 256 and linearly scaling
        it down/up for other batch sizes. The weight decay is set to 1/32768 for all
        parameters except the batch normalization layers. For further information check:
        https://catalog.ngc.nvidia.com/orgs/nvidia/resources/resnet_50_v1_5_for_pytorch

        Returns
        -------
        List
            Returns the optimizer and the learning rate scheduler

        Raises
        ------
        ValueError
            If the scheduler is not implemented
        """

        optimizer_params = [
            {
                "params": [
                    param
                    for name, param in self.named_parameters()
                    if not any(nd in name for nd in ["bn"]) and name != "log_std"
                ],
                "weight_decay": 1 / 32768,  # Include weight decay for other parameters
            },
            {
                "params": [
                    param
                    for name, param in self.named_parameters()
                    if any(nd in name for nd in ["bn"])
                ],
                "weight_decay": 0.0,  # Exclude weight decay for parameters with 'bn' in name
            },
        ]

        # Conditionally add parameter group for log_std based on self.loss_type
        if self.loss_type == "nll":
            optimizer_params.append(
                {
                    "params": [self.log_std],  # Separate parameter group for log_std
                    "weight_decay": 0.0,  # Exclude weight decay for log_std
                }
            )

        if self.optimizer == "SGD":
            optimizer = torch.optim.SGD(
                optimizer_params,
                lr=self.set_lr,
                momentum=0.875,
                nesterov=True,
            )
        elif self.optimizer == "AdamW":
            optimizer = torch.optim.AdamW(
                optimizer_params,
                lr=self.set_lr,
                betas=(0.9, 0.999),
                eps=1e-08,
            )
        else:
            raise NotImplementedError("Optimizer has not been implemented")

        if self.config_scheduler == "CosineAnnealingWarmRestarts":
            lr_scheduler = {
                "scheduler": CosineAnnealingWarmRestarts(
                    optimizer, T_0=5, eta_min=1e-4
                ),
                "monitor": self.monitor,
                "interval": "epoch",
            }

        elif self.config_scheduler == "OneCycleLR":
            lr_scheduler = {
                "scheduler": OneCycleLR(
                    optimizer,
                    max_lr=0.01,
                    total_steps=self.trainer.estimated_stepping_batches,
                ),
                "monitor": self.monitor,
                "interval": "step",
            }
        elif self.config_scheduler == "LinearWarmupCosineAnnealingLR":
            num_epochs = self.trainer.max_epochs
            total_steps = self.trainer.estimated_stepping_batches
            steps_per_epoch = total_steps // num_epochs
            # Warmup for 5% of the total steps
            warmup_steps = steps_per_epoch * int(num_epochs * 0.05)

            def lr_lambda(current_step):
                if current_step < warmup_steps:
                    # Linear warmup phase
                    return current_step / max(1, warmup_steps)
                else:
                    # Cosine annealing phase
                    eta_min = 1e-5
                    remaining_steps = current_step - warmup_steps
                    current_epoch = remaining_steps // steps_per_epoch
                    cosine_factor = 0.5 * (
                        1 + math.cos(math.pi * current_epoch / num_epochs)
                    )
                    return eta_min + (1 - eta_min) * cosine_factor

            lr_scheduler = {
                "scheduler": LambdaLR(optimizer, lr_lambda=lr_lambda),
                "monitor": self.monitor,
                "interval": "step",
            }
        elif self.config_scheduler == "CosineAnnealingLR":
            num_epochs = self.trainer.max_epochs
            lr_scheduler = {
                "scheduler": CosineAnnealingLR(
                    optimizer,
                    T_max=num_epochs,
                    eta_min=1e-4,
                ),
                "monitor": self.monitor,
                "interval": "epoch",
            }
        else:
            raise ValueError(f"{self.config_scheduler} lr_scheduler is not implemented")

        return [optimizer], [lr_scheduler]

    def symmetrize(self, data_reconstructed: torch.Tensor) -> torch.Tensor:
        """
        Symmetrizes the reconstructed data so that the weights can learn other patterns.
        Loss calculated only on the reconstruction faithfulness of the upper triangle
        of the distance map

        Parameters
        ----------
        data_reconstructed : torch.Tensor
            Reconstructed data; output of the decoder

        Returns
        -------
        torch.Tensor
            Symmetric version of the reconstructed data
        """
        # Get the upper triangular part of each tensor in the batch
        upper_triangles = torch.triu(data_reconstructed)

        # Symmetrize each tensor in the batch individually
        symmetrized_arrays = upper_triangles + torch.transpose(upper_triangles, -1, -2)

        # Fill diagonal elements with zeros for each tensor individually
        diag_values = torch.diagonal(symmetrized_arrays, dim1=-2, dim2=-1)
        symmetrized_arrays = symmetrized_arrays - torch.diag_embed(diag_values)

        return symmetrized_arrays

    def sample(self, num_samples: int, sequence: torch.Tensor = None) -> torch.Tensor:
        """
        Sample from the latent space of the VAE and optionally
        condition on a sequence

        Parameters
        ----------
        num_samples : int
            Number of samples to generate
        sequence : torch.Tensor
            Sequence to generate distance maps for (sequence needs to be < 385), by default None

        Returns
        -------
        torch.Tensor
            Returns the generated distance maps
        """
        # If no sequence is given, generate zeroes (no conditioning)
        if sequence is None:
            sequence = torch.zeros(
                (num_samples, self.hparams.dimension, self.vocab_size),
                dtype=torch.float32,
            ).to(self.device)
        else:
            sequence = [sequence] * num_samples
        # Sample the latent encoding from N(0, I)
        latent_samples = torch.randn(
            num_samples, 1, int(self.compressed_size), int(self.compressed_size)
        ).to(self.device)

        # Decode the samples conditioned on sequence/labels
        with torch.no_grad():
            generated_samples = self.decode(latent_samples, sequence)

        if sequence is not None:
            sequence_length = len(sequence[0])
            generated_samples = generated_samples[
                :, :, :sequence_length, :sequence_length
            ]
        return generated_samples



import torch.nn.functional as F
from torch import nn

from starling.models.blocks import (
    ResBlockDecBasic,
    ResBlockDecBottleneck,
    ResBlockEncBasic,
    ResBlockEncBottleneck,
)


class ResNet_Encoder_Original(nn.Module):
    def __init__(
        self,
        in_channels,
        num_blocks,
        kernel_size=None,
        dimension=None,
        block_type=ResBlockEncBasic,
        base=64,
    ) -> None:
        super().__init__()

        self.block_type = block_type

        # First convolution of the ResNet Encoder reduction in the spatial dimensions / 2
        # with kernel=7 and stride=2 AvgPool2d reduces spatial dimensions by / 2
        self.in_channels = 64
        self.first_conv = nn.Sequential(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=self.in_channels,
                kernel_size=7,
                stride=2,
                padding=3,
            ),
            nn.BatchNorm2d(self.in_channels),
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
        )

        self.layer1 = self._make_layer(self.block_type, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(self.block_type, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(self.block_type, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(self.block_type, 512, num_blocks[3], stride=2)

        self.average_pool = nn.AdaptiveAvgPool2d((1, 1))

    def _make_layer(self, block, out_channels, blocks, stride=1):
        layers = []
        layers.append(block(self.in_channels, out_channels, stride))
        self.in_channels = out_channels * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.in_channels, out_channels, stride=1))
        return nn.Sequential(*layers)

    def forward(self, data):
        data = self.first_conv(data)
        # for layer in self.layers:
        #     data = layer(data)
        # The final adaptive average can also be done through convolution
        data = self.layer1(data)
        data = self.layer2(data)
        data = self.layer3(data)
        data = self.layer4(data)
        data = self.average_pool(data)
        return data


class ResNet_Decoder_Original(nn.Module):
    def __init__(
        self,
        out_channels,
        num_blocks,
        kernel_size,
        dimension,
        block_type=ResBlockDecBasic,
        base=64,
    ) -> None:
        super().__init__()

        # Calculate the input channels from the encoder, assuming
        # symmetric encoder and decoder setup
        self.block_type = block_type
        if self.block_type == ResBlockDecBasic:
            self.in_channels = 512
        else:
            self.in_channels = 4096

        self.interpolate = int(dimension / (2 ** (len(num_blocks) + 1)))

        # This part can be done in many ways, this is just one of them
        # It adds some number of parameters
        # self.resize_conv = ResizeConv2d(
        #     in_channels=self.in_channels,
        #     out_channels=self.in_channels,
        #     kernel_size=kernel_size,
        #     size=(self.interpolate, self.interpolate),
        #     mode="nearest",
        # )

        self.layers = nn.ModuleList()

        self.layer1 = self._make_layer(self.block_type, 512, num_blocks[0], stride=2)
        self.layer2 = self._make_layer(self.block_type, 256, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(self.block_type, 128, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(
            self.block_type, 64, num_blocks[3], stride=1, last_layer=True
        )

        # for layer in [self.layer1, self.layer2, self.layer3, self.layer4]:
        #     self.layers.append(layer)

        # # This part could be done through interpolation (analogous to MaxPool)
        self.reshaping_conv = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels=64,
                out_channels=64,
                kernel_size=kernel_size,
                stride=2,
                padding=1,
                output_padding=1,
            ),
            nn.BatchNorm2d(64),
            # nn.LayerNorm([64, int(dimension / 2), int(dimension / 2)]),
            nn.ReLU(inplace=True),
        )

        # Final output layer that looks similar to the first layer of
        # the ResNet Encoder
        self.output_layer = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels=64,
                out_channels=out_channels,
                kernel_size=7,
                stride=2,
                padding=3,
                output_padding=1,
            ),
            nn.ReLU(inplace=True),
        )

    def _make_layer(self, block, out_channels, blocks, stride=1, last_layer=False):
        layers = []
        self.in_channels = out_channels * block.contraction
        for _ in range(1, blocks):
            layers.append(block(self.in_channels, out_channels, stride=1))
        if stride > 1 and block == ResBlockDecBasic:
            out_channels = int(out_channels / 2)
        layers.append(
            block(self.in_channels, out_channels, stride, last_layer=last_layer)
        )
        return nn.Sequential(*layers)

    def forward(self, data):
        # data = self.resize_conv(data)
        data = F.interpolate(data, size=(self.interpolate, self.interpolate))
        data = self.layer1(data)
        data = self.layer2(data)
        data = self.layer3(data)
        data = self.layer4(data)
        data = self.reshaping_conv(data)
        data = self.output_layer(data)
        return data


# Current implementations of ResNets


def Resnet18_Encoder(in_channels, kernel_size, dimension, base):
    return ResNet_Encoder_Original(
        block_type=ResBlockEncBasic,
        in_channels=in_channels,
        num_blocks=[2, 2, 2, 2],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )


def Resnet18_Decoder(out_channels, kernel_size, dimension, base):
    return ResNet_Decoder_Original(
        block_type=ResBlockDecBasic,
        out_channels=out_channels,
        num_blocks=[2, 2, 2, 2],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )


def Resnet34_Encoder(in_channels, kernel_size, dimension, base):
    return ResNet_Encoder_Original(
        block_type=ResBlockEncBasic,
        in_channels=in_channels,
        num_blocks=[3, 4, 6, 3],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )


def Resnet34_Decoder(out_channels, kernel_size, dimension, base):
    return ResNet_Decoder_Original(
        block_type=ResBlockDecBasic,
        out_channels=out_channels,
        num_blocks=[3, 6, 4, 3],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )


def Resnet50_Encoder(in_channels, kernel_size, dimension, base):
    return ResNet_Encoder_Original(
        block_type=ResBlockEncBottleneck,
        in_channels=in_channels,
        num_blocks=[3, 4, 6, 3],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )


def Resnet50_Decoder(out_channels, kernel_size, dimension, base):
    return ResNet_Decoder_Original(
        block_type=ResBlockDecBottleneck,
        out_channels=out_channels,
        num_blocks=[3, 6, 4, 3],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )


def Resnet101_Encoder(in_channels, kernel_size, dimension, base):
    return ResNet_Encoder_Original(
        block_type=ResBlockEncBottleneck,
        in_channels=in_channels,
        num_blocks=[3, 4, 23, 3],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )


def Resnet101_Decoder(out_channels, kernel_size, dimension, base):
    return ResNet_Decoder_Original(
        block_type=ResBlockDecBottleneck,
        out_channels=out_channels,
        num_blocks=[3, 23, 4, 3],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )


def Resnet152_Encoder(in_channels, kernel_size, dimension, base):
    return ResNet_Encoder_Original(
        block_type=ResBlockEncBottleneck,
        in_channels=in_channels,
        num_blocks=[3, 8, 36, 3],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )


def Resnet152_Decoder(out_channels, kernel_size, dimension, base):
    return ResNet_Decoder_Original(
        block_type=ResBlockDecBottleneck,
        out_channels=out_channels,
        num_blocks=[3, 36, 8, 3],
        kernel_size=kernel_size,
        dimension=dimension,
        base=base,
    )



import os

import torch

from starling import configs

# local imports
from starling.configs import DEFAULT_DDPM_WEIGHTS_PATH, DEFAULT_ENCODER_WEIGHTS_PATH
from starling.models.diffusion import DiffusionModel
from starling.models.unet import UNetConditional
from starling.models.vae import VAE


class ModelManager:
    def __init__(self):
        self.encoder_model = None
        self.diffusion_model = None

    def load_models(self, encoder_path, ddpm_path, device):
        """Load the models from local files or URLs."""

        def load_from_path_or_url(path):
            if path.startswith("http"):
                # Download from URL if not cached
                cache_dir = torch.hub.get_dir() + "/checkpoints/"
                os.makedirs(cache_dir, exist_ok=True)
                cached_path = cache_dir + os.path.basename(path)
                if not os.path.exists(cached_path):
                    torch.hub.download_url_to_file(path, cached_path)
                return cached_path
            return path

        # Resolve paths
        encoder_path = load_from_path_or_url(encoder_path)
        ddpm_path = load_from_path_or_url(ddpm_path)

        # Continue with existing loading logic
        if not os.path.exists(encoder_path):
            raise FileNotFoundError(f"Encoder model {encoder_path} not found.")
        if not os.path.exists(ddpm_path):
            raise FileNotFoundError(f"DDPM model {ddpm_path} not found.")

        # Load the encoder model
        encoder_model = VAE.load_from_checkpoint(encoder_path, map_location=device)

        # Load the diffusion model
        diffusion_model = DiffusionModel.load_from_checkpoint(
            ddpm_path,
            model=UNetConditional(
                in_channels=1,
                out_channels=1,
                base=64,
                norm="group",
                blocks=[2, 2, 2],
                middle_blocks=2,
                labels_dim=configs.UNET_LABELS_DIM,
            ),
            encoder_model=encoder_model,
            map_location=device,
        )

        return encoder_model, diffusion_model

    def get_models(
        self,
        encoder_path=DEFAULT_ENCODER_WEIGHTS_PATH,
        ddpm_path=DEFAULT_DDPM_WEIGHTS_PATH,
        device="cpu",
    ):
        """
        Lazy-load models if not already loaded.

        Parameters
        ----------
        encoder_path : str
            The path to the encoder model.
            Default is
        ddpm_path : str
            The path to the DDPM model.
        device : str
            The device on which to load the models. Default is CPU,
            but this changes depending on whatever we want to use
            in ensemble_generation.py. Just made CPU default because
            all platforms have CPU.

        Returns
        -------
        encoder_model, diffusion_model
            The loaded encoder and diffusion models.
        """
        if self.encoder_model is None or self.diffusion_model is None:
            # Models haven't been loaded yet, so load them now
            self.encoder_model, self.diffusion_model = self.load_models(
                encoder_path, ddpm_path, device
            )

        # Return the already-loaded models
        return self.encoder_model, self.diffusion_model



import math
from typing import List

import torch
from torch import nn

from starling.models.blocks import ResBlockEncBasic, ResizeConv2d
from starling.models.normalization import RMSNorm
from starling.models.transformer import SpatialTransformer


class SinusoidalPosEmb(nn.Module):
    def __init__(self, dim: int, theta: int = 10000):
        """
        Generates sinusoidal positional embeddings that are used in the denoising-diffusion
        models to encode the timestep information. The positional embeddings are generated
        using sine and cosine functions. It takes in time in the shape of (batch_size, 1)
        and returns the positional embeddings in the shape of (batch_size, dim). The positional
        encodings are later used in each of the ResNet blocks to encode the timestep information.

        Parameters
        ----------
        dim : int
            Dimension of the input data.
        theta : int, optional
            A scaling factor for the positional embeddings. The default value is 10000.
        """
        super().__init__()
        self.dim = dim
        self.theta = theta

    def forward(self, time: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the positional (timestep) embeddings.

        Parameters
        ----------
        time : torch.Tensor
            Timestep information in the shape of (batch_size, 1).

        Returns
        -------
        torch.Tensor
            Positional (timestep) embeddings in the shape of (batch_size, dim).
        """
        device = time.device

        # The number of unique frequencies in the positional embeddings, half
        # will be used for sine and the other half for cosine functions
        half_dim = self.dim // 2
        emb = math.log(self.theta) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, device=device) * -emb)
        emb = time[:, None] * emb[None, :]
        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)
        return emb


class ConditionalSequential(nn.Sequential):
    def forward(self, x, condition):
        for module in self._modules.values():
            x = module(x, condition)
        return x


class Downsample(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, norm: str):
        """
        A convolutional block that reduces the spatial dimensions of the input tensor by a factor of 2.
        The block consists of a convolutional layer with a kernel size of 3, stride of 2, and padding of 1.
        The convolutional layer is followed by a normalization layer and a ReLU activation function.

        Parameters
        ----------
        in_channels : int
            The number of features in the input tensor.
        out_channels : int
            The number of features in the output tensor.
        norm : str
            The normalization layer to be used in the block. Choose from batch, instance, rms, or group.
        """
        super().__init__()

        normalization = {
            "batch": nn.BatchNorm2d,
            "instance": nn.InstanceNorm2d,
            "rms": RMSNorm,
            "group": nn.GroupNorm,
        }

        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=2, padding=1),
            normalization[norm](out_channels)
            if norm != "group"
            else normalization[norm](32, out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        return x


class ResnetLayer(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        norm,
        num_blocks,
        timestep_dim,
        class_dim=None,
    ):
        super().__init__()

        self.layer = nn.ModuleList()

        self.in_channels = in_channels

        for block in range(num_blocks):
            self.layer.append(
                ResBlockEncBasic(
                    self.in_channels, out_channels, 1, norm, timestep_dim, class_dim
                )
            )

            self.in_channels = out_channels

    def forward(self, x, time):
        for layer in self.layer:
            x = layer(x, time)
        return x


class CrossAttentionResnetLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        norm: str,
        num_blocks: int,
        attention_heads: int,
        timestep_dim: int,
        label_dim: int,
        custom_attention: bool,
    ):
        """
        A combination of ResNet blocks followed by spatial transformer blocks. The ResNet block
        processes the input tensor and the spatial transformer block captures the relationships
        between the input tensor and the context data (protein sequences).

        Parameters
        ----------
        in_channels : int
            The number of features in the input tensor.
        out_channels : int
            The number of features in the output tensor.
        norm : str
            The normalization layer to be used in the block. Choose from batch, instance, rms, or group.
        num_blocks : int
            The number of ResNet + spatial transformer blocks in the layer.
        attention_heads : int
            The number of heads in the multi-head attention layer.
        timestep_dim : int
            The dimension of the timestep embeddings.
        label_dim : int
            The dimension of the context data (protein sequences).
        """
        super().__init__()

        self.layer = nn.ModuleList()
        self.transformer = nn.ModuleList()

        self.in_channels = in_channels

        for block in range(num_blocks):
            self.layer.append(
                ResBlockEncBasic(self.in_channels, out_channels, 1, norm, timestep_dim)
            )
            self.transformer.append(
                SpatialTransformer(
                    out_channels, attention_heads, label_dim, custom_attention
                ),
            )

            self.in_channels = out_channels

    def forward(
        self, x: torch.Tensor, time: torch.Tensor, sequence_label: torch.Tensor
    ) -> torch.Tensor:
        """
        Forward pass of the ResNet + spatial transformer blocks.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor to be processed by the ResNet + spatial transformer blocks.
        time : torch.Tensor
            Timestep embeddings to be used by the network.
        sequence_label : torch.Tensor
            Context data (protein sequences) to guide the prediction.

        Returns
        -------
        torch.Tensor
            Output of the ResNet + spatial transformer blocks.
        """
        for layer, transformer in zip(self.layer, self.transformer):
            x = layer(x, time)
            x = transformer(x, context=sequence_label)
        return x


class UNetConditional(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        base: int,
        norm: str,
        blocks: List = [2, 2, 2],
        middle_blocks: int = 2,
        labels_dim: int = 512,
        sinusoidal_pos_emb_theta: int = 10000,
        custom_attention: bool = True,
    ):
        """
        A U-Net architecture that uses ResNet blocks with spatial transformer blocks to process the input
        tensor and the context data (protein sequences). The U-Net architecture consists of an encoder,
        a middle section, and a decoder. The spatial transformer blocks are used to capture the relationships
        between the input tensor and the context data (in our case protein sequences).

        Parameters
        ----------
        in_channels : int
            The number of features in the input tensor.
        out_channels : int
            The number of features in the output tensor.
        base : int
            The base number of features in the U-Net architecture.
        norm : str
            The normalization layer to be used in the block. Choose from batch, instance, rms, or group
        blocks : List, optional
            The number of ResNet + spatial transformer blocks in each section of the U-Net architecture, by default [2, 2, 2]
        middle_blocks : int, optional
            The number of ResNet + spatial transformer blocks in the middle section of the U-Net architecture, by default 2
        labels_dim : int, optional
            The dimension of the context data (i.e., protein sequences), by default 512
        sinusoidal_pos_emb_theta : int, optional
            A scaling factor for the positional (timestep) embeddings, by default 10000
        """
        super().__init__()

        normalization = {
            "batch": nn.BatchNorm2d,
            "instance": nn.InstanceNorm2d,
            "rms": RMSNorm,
            "group": nn.GroupNorm,
        }

        self.norm = norm
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.time_dim = base * 4
        self.base = base
        self.labels_dim = labels_dim

        # Time embeddings
        self.time_emb = SinusoidalPosEmb(self.base, theta=sinusoidal_pos_emb_theta)
        self.time_mlp = nn.Sequential(
            self.time_emb,
            nn.Linear(self.base, self.time_dim),
            nn.SiLU(inplace=False),
            nn.Linear(self.time_dim, self.time_dim),
        )

        all_in_channels = [base * (2**i) for i in range(len(blocks) + 1)]

        # Encoder part of UNet

        self.conv_in = CrossAttentionResnetLayer(
            in_channels,
            all_in_channels[0],
            self.norm,
            blocks[0],
            8,
            self.time_dim,
            self.labels_dim,
            custom_attention,
        )

        self.encoder_layer1 = CrossAttentionResnetLayer(
            all_in_channels[0],
            all_in_channels[0],
            self.norm,
            blocks[0],
            8,
            self.time_dim,
            self.labels_dim,
            custom_attention,
        )

        self.downsample1 = Downsample(all_in_channels[0], all_in_channels[1], norm)

        self.encoder_layer2 = CrossAttentionResnetLayer(
            all_in_channels[1],
            all_in_channels[1],
            self.norm,
            blocks[1],
            8,
            self.time_dim,
            self.labels_dim,
            custom_attention,
        )

        self.downsample2 = Downsample(all_in_channels[1], all_in_channels[2], norm)

        self.encoder_layer3 = CrossAttentionResnetLayer(
            all_in_channels[2],
            all_in_channels[2],
            self.norm,
            blocks[2],
            8,
            self.time_dim,
            self.labels_dim,
            custom_attention,
        )

        self.downsample3 = Downsample(all_in_channels[2], all_in_channels[3], norm)

        # Middle convolution of the UNet

        self.middle = CrossAttentionResnetLayer(
            all_in_channels[3],
            all_in_channels[3],
            self.norm,
            middle_blocks,
            8,
            self.time_dim,
            self.labels_dim,
            custom_attention,
        )

        # Decoder part of UNet

        self.upconv1 = ResizeConv2d(
            all_in_channels[3],
            all_in_channels[2],
            kernel_size=3,
            padding=1,
            scale_factor=2,
            norm=normalization[norm],
            activation="relu",
        )

        self.decoder_layer1 = CrossAttentionResnetLayer(
            all_in_channels[2] * 2,
            all_in_channels[2],
            self.norm,
            blocks[2],
            8,
            self.time_dim,
            self.labels_dim,
            custom_attention,
        )

        self.upconv2 = ResizeConv2d(
            all_in_channels[2],
            all_in_channels[1],
            kernel_size=3,
            padding=1,
            scale_factor=2,
            norm=normalization[norm],
            activation="relu",
        )

        self.decoder_layer2 = CrossAttentionResnetLayer(
            all_in_channels[1] * 2,
            all_in_channels[1],
            self.norm,
            blocks[1],
            8,
            self.time_dim,
            self.labels_dim,
            custom_attention,
        )

        self.upconv3 = ResizeConv2d(
            all_in_channels[1],
            all_in_channels[0],
            kernel_size=3,
            padding=1,
            scale_factor=2,
            norm=normalization[norm],
            activation="relu",
        )

        self.decoder_layer3 = CrossAttentionResnetLayer(
            all_in_channels[0] * 2,
            all_in_channels[0],
            self.norm,
            blocks[1],
            8,
            self.time_dim,
            self.labels_dim,
            custom_attention,
        )

        self.conv_out = nn.Conv2d(all_in_channels[0], out_channels, kernel_size=1)

    def forward(
        self, x: torch.Tensor, time: torch.Tensor, labels: torch.Tensor = None
    ) -> torch.Tensor:
        """
        Forward pass of the UNet architecture.

        Parameters
        ----------
        x : torch.Tensor
            Data to pass through the UNet architecture.
        time : torch.Tensor
            Timestep embeddings.
        labels : torch.Tensor, optional
            Context data (protein sequences) to guide the prediction, by default None

        Returns
        -------
        torch.Tensor
            Output of the UNet architecture.
        """
        # Get the time embeddings
        time = self.time_mlp(time)

        # Initial convolution
        x = self.conv_in(x, time, labels)

        # Encoder forward passes
        x = self.encoder_layer1(x, time, labels)
        x_layer1 = x.clone()
        x = self.downsample1(x)

        x = self.encoder_layer2(x, time, labels)
        x_layer2 = x.clone()
        x = self.downsample2(x)

        x = self.encoder_layer3(x, time, labels)
        x_layer3 = x.clone()
        x = self.downsample3(x)

        # Mid UNet
        x = self.middle(x, time, labels)

        # Decoder forward passes with skip connections from the encoder
        x = self.upconv1(x)
        x = torch.cat((x, x_layer3), dim=1)
        x = self.decoder_layer1(x, time, labels)

        x = self.upconv2(x)
        x = torch.cat((x, x_layer2), dim=1)
        x = self.decoder_layer2(x, time, labels)

        x = self.upconv3(x)
        x = torch.cat((x, x_layer1), dim=1)
        x = self.decoder_layer3(x, time, labels)

        # Final convolutions
        x = self.conv_out(x)

        return x



from typing import List

from torch import nn

from starling.models.blocks import (
    LayerNorm,
    ResBlockDecBasic,
    ResBlockEncBasic,
    ResizeConv2d,
)


class ResNet_Encoder(nn.Module):
    def __init__(
        self,
        in_channels,
        num_blocks,
        norm,
        base=64,
        block_type=ResBlockEncBasic,
    ) -> None:
        super().__init__()

        self.block_type = block_type
        self.norm = norm
        normalization = {
            "batch": nn.BatchNorm2d,
            "instance": nn.InstanceNorm2d,
            "layer": LayerNorm,
            "group": nn.GroupNorm,
        }

        # First convolution of the ResNet Encoder reduction in the spatial dimensions / 2
        # with kernel=7 and stride=2 AvgPool2d reduces spatial dimensions by / 2
        self.first_conv = nn.Sequential(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=base,
                kernel_size=7,
                stride=2,
                padding=3,
            ),
            normalization[norm](base)
            if norm != "group"
            else normalization[norm](32, base),
        )

        self.in_channels = base

        layer_in_channels = [base * (2**i) for i in range(len(num_blocks))]

        # Setting up the layers for the encoder
        self.layer1 = self._make_layer(
            self.block_type, layer_in_channels[0], num_blocks[0], stride=1
        )
        self.layer2 = self._make_layer(
            self.block_type, layer_in_channels[1], num_blocks[1], stride=2
        )
        self.layer3 = self._make_layer(
            self.block_type, layer_in_channels[2], num_blocks[2], stride=2
        )
        self.layer4 = self._make_layer(
            self.block_type, layer_in_channels[3], num_blocks[3], stride=2
        )

    def _make_layer(self, block, out_channels, blocks, stride=1):
        layers = nn.ModuleList()
        # layers = []
        layers.append(
            block(
                self.in_channels,
                out_channels,
                stride,
                norm=self.norm,
            )
        )
        self.in_channels = out_channels * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.in_channels,
                    out_channels,
                    stride=1,
                    norm=self.norm,
                )
            )
        return layers
        # return nn.Sequential(*layers)

    def forward(self, data):
        data = self.first_conv(data)

        for block in self.layer1:
            data = block(data)

        for block in self.layer2:
            data = block(data)

        for block in self.layer3:
            data = block(data)

        for block in self.layer4:
            data = block(data)

        return data


class ResNet_Decoder(nn.Module):
    def __init__(
        self,
        out_channels: int,
        num_blocks: List,
        dimension: int,
        norm: str,
        block_type=ResBlockDecBasic,
        base=64,
    ) -> None:
        super().__init__()

        self.norm = norm

        # Calculate the input channels from the encoder, assuming
        # symmetric encoder and decoder setup
        self.block_type = block_type
        if self.block_type == ResBlockDecBasic:
            layer_in_channels = [base * (2**i) for i in range(len(num_blocks))]
            self.in_channels = layer_in_channels[-1]
        else:
            layer_in_channels = [base * (4**i) for i in range(len(num_blocks))]
            self.in_channels = layer_in_channels[-1]

        # Setting up the layers for the decoder

        self.layer1 = self._make_layer(
            self.block_type, layer_in_channels[-1], num_blocks[0], stride=2
        )
        self.layer2 = self._make_layer(
            self.block_type, layer_in_channels[-2], num_blocks[1], stride=2
        )
        self.layer3 = self._make_layer(
            self.block_type, layer_in_channels[-3], num_blocks[2], stride=2
        )
        self.layer4 = self._make_layer(
            self.block_type,
            layer_in_channels[-4],
            num_blocks[3],
            stride=1,
            last_layer=True,
        )

        in_channels_post_resnets = layer_in_channels[-4]

        self.output_layer = ResizeConv2d(
            in_channels=in_channels_post_resnets,
            out_channels=out_channels,
            kernel_size=7,
            padding=3,
            norm=None,
            activation="relu",
            scale_factor=2,
        )

    def _make_layer(self, block, out_channels, blocks, stride=1, last_layer=False):
        layers = nn.ModuleList()
        self.in_channels = out_channels * block.contraction
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.in_channels,
                    out_channels,
                    stride=1,
                    norm=self.norm,
                )
            )
        if stride > 1 and block == ResBlockDecBasic:
            out_channels = int(out_channels / 2)
        layers.append(
            block(
                self.in_channels,
                out_channels,
                stride,
                last_layer=last_layer,
                norm=self.norm,
            )
        )

        return layers

    def forward(self, data):
        for block in self.layer1:
            data = block(data)

        for block in self.layer2:
            data = block(data)

        for block in self.layer3:
            data = block(data)

        for block in self.layer4:
            data = block(data)

        data = self.output_layer(data)
        return data


class ConditionalSequential(nn.Sequential):
    def forward(self, x, condition=None):
        if condition is None:
            for module in self._modules.values():
                x = module(x)
        else:
            for module in self._modules.values():
                x = module(x, condition)
        return x


# Current implementations of ResNets


def Resnet18_Encoder(in_channels, norm, base):
    return ResNet_Encoder(
        in_channels,
        num_blocks=[2, 2, 2, 2],
        base=base,
        norm=norm,
    )


def Resnet18_Decoder(out_channels, dimension, base, norm):
    return ResNet_Decoder(
        out_channels,
        num_blocks=[2, 2, 2, 2],
        dimension=dimension,
        base=base,
        norm=norm,
    )


def Resnet34_Encoder(in_channels, base, norm):
    return ResNet_Encoder(
        in_channels,
        num_blocks=[3, 4, 6, 3],
        base=base,
        norm=norm,
    )


def Resnet34_Decoder(out_channels, dimension, base, norm):
    return ResNet_Decoder(
        out_channels,
        num_blocks=[3, 4, 6, 3],
        dimension=dimension,
        base=base,
        norm=norm,
    )



import numpy as np
import torch
from einops import rearrange
from torch import nn

# Adapted from https://github.com/CompVis/taming-transformers/blob/master/taming/modules/vqvae/quantize.py


class VectorQuantizer2(nn.Module):
    """
    Improved version over VectorQuantizer, can be used as a drop-in replacement. Mostly
    avoids costly matrix multiplications and allows for post-hoc remapping of indices.
    """

    # NOTE: due to a bug the beta term was applied to the wrong term. for
    # backwards compatibility we use the buggy version by default, but you can
    # specify legacy=False to fix it.
    def __init__(
        self,
        n_e,
        e_dim,
        beta,
        remap=None,
        unknown_index="random",
        sane_index_shape=False,
        legacy=True,
    ):
        super().__init__()
        self.n_e = n_e
        self.e_dim = e_dim
        self.beta = beta
        self.legacy = legacy

        self.embedding = nn.Embedding(self.n_e, self.e_dim)
        self.embedding.weight.data.uniform_(-1.0 / self.n_e, 1.0 / self.n_e)

        self.remap = remap
        if self.remap is not None:
            self.register_buffer("used", torch.tensor(np.load(self.remap)))
            self.re_embed = self.used.shape[0]
            self.unknown_index = unknown_index  # "random" or "extra" or integer
            if self.unknown_index == "extra":
                self.unknown_index = self.re_embed
                self.re_embed = self.re_embed + 1
            print(
                f"Remapping {self.n_e} indices to {self.re_embed} indices. "
                f"Using {self.unknown_index} for unknown indices."
            )
        else:
            self.re_embed = n_e

        self.sane_index_shape = sane_index_shape

    def remap_to_used(self, inds):
        ishape = inds.shape
        assert len(ishape) > 1
        inds = inds.reshape(ishape[0], -1)
        used = self.used.to(inds)
        match = (inds[:, :, None] == used[None, None, ...]).long()
        new = match.argmax(-1)
        unknown = match.sum(2) < 1
        if self.unknown_index == "random":
            new[unknown] = torch.randint(0, self.re_embed, size=new[unknown].shape).to(
                device=new.device
            )
        else:
            new[unknown] = self.unknown_index
        return new.reshape(ishape)

    def unmap_to_all(self, inds):
        ishape = inds.shape
        assert len(ishape) > 1
        inds = inds.reshape(ishape[0], -1)
        used = self.used.to(inds)
        if self.re_embed > self.used.shape[0]:  # extra token
            inds[inds >= self.used.shape[0]] = 0  # simply set to zero
        back = torch.gather(used[None, :][inds.shape[0] * [0], :], 1, inds)
        return back.reshape(ishape)

    def forward(self, z, temp=None, rescale_logits=False, return_logits=False):
        assert temp is None or temp == 1.0, "Only for interface compatible with Gumbel"
        assert rescale_logits == False, "Only for interface compatible with Gumbel"
        assert return_logits == False, "Only for interface compatible with Gumbel"
        # reshape z -> (batch, height, width, channel) and flatten
        z = rearrange(z, "b c h w -> b h w c").contiguous()
        z_flattened = z.view(-1, self.e_dim)
        # distances from z to embeddings e_j (z - e)^2 = z^2 + e^2 - 2 e * z

        d = (
            torch.sum(z_flattened**2, dim=1, keepdim=True)
            + torch.sum(self.embedding.weight**2, dim=1)
            - 2
            * torch.einsum(
                "bd,dn->bn", z_flattened, rearrange(self.embedding.weight, "n d -> d n")
            )
        )

        min_encoding_indices = torch.argmin(d, dim=1)
        z_q = self.embedding(min_encoding_indices).view(z.shape)
        perplexity = None
        min_encodings = None

        # compute loss for embedding
        if not self.legacy:
            loss = self.beta * torch.mean((z_q.detach() - z) ** 2) + torch.mean(
                (z_q - z.detach()) ** 2
            )
        else:
            loss = torch.mean((z_q.detach() - z) ** 2) + self.beta * torch.mean(
                (z_q - z.detach()) ** 2
            )

        # preserve gradients
        z_q = z + (z_q - z).detach()

        # reshape back to match original input shape
        z_q = rearrange(z_q, "b h w c -> b c h w").contiguous()

        if self.remap is not None:
            min_encoding_indices = min_encoding_indices.reshape(
                z.shape[0], -1
            )  # add batch axis
            min_encoding_indices = self.remap_to_used(min_encoding_indices)
            min_encoding_indices = min_encoding_indices.reshape(-1, 1)  # flatten

        if self.sane_index_shape:
            min_encoding_indices = min_encoding_indices.reshape(
                z_q.shape[0], z_q.shape[2], z_q.shape[3]
            )

        return z_q, loss, (perplexity, min_encodings, min_encoding_indices)

    def get_codebook_entry(self, indices, shape):
        # shape specifying (batch, height, width, channel)
        if self.remap is not None:
            indices = indices.reshape(shape[0], -1)  # add batch axis
            indices = self.unmap_to_all(indices)
            indices = indices.reshape(-1)  # flatten again

        # get quantized latent vectors
        z_q = self.embedding(indices)

        if shape is not None:
            z_q = z_q.view(shape)
            # reshape back to match original input shape
            z_q = z_q.permute(0, 3, 1, 2).contiguous()

        return z_q



import argparse
import os
import sys
import time
from argparse import ArgumentParser

import numpy as np
import psutil

from starling import configs
from starling._version import __version__
from starling.frontend.ensemble_generation import generate
from starling.utilities import check_device, check_file_exists


def print_starling():
    print("\n-------------------------------------------------------")
    print(f"     STARLING (version {__version__})           ")
    print("-------------------------------------------------------")


def print_info():
    # local imports if needed
    print_starling()
    print("Using models at the following locations:")
    print(f"  VAE  model weights: {configs.DEFAULT_ENCODER_WEIGHTS_PATH}")
    print(f"  DDPM model weights: {configs.DEFAULT_DDPM_WEIGHTS_PATH}")
    print("-------------------------------------------------------")
    print("CONFIG INFO:")
    print("  Default # of confs :", configs.DEFAULT_NUMBER_CONFS)
    print("  Default batch size :", configs.DEFAULT_BATCH_SIZE)
    print("  Default steps      :", configs.DEFAULT_STEPS)
    print("  Default # of CPUs  :", configs.DEFAULT_CPU_COUNT_MDS)
    print("  Default # MDS jobs :", configs.DEFAULT_MDS_NUM_INIT)
    print(f"  Default device     : {check_device(None)}")
    print("-------------------------------------------------------")
    print(
        "Need help - please raise an issue on GitHub: https://github.com/idptools/starling/issues"
    )

    print("\n")


def main():
    # Initialize the argument parser
    parser = ArgumentParser(description="Generate distance maps using STARLING.")

    # Add command-line arguments corresponding to the parameters of the generate function
    parser.add_argument(
        "user_input",
        type=str,
        help="Input sequences in various formats (file, string, list, or dict)",
        nargs="?",
        default=None,
    )
    parser.add_argument(
        "-c",
        "--conformations",
        type=int,
        default=configs.DEFAULT_NUMBER_CONFS,
        help=f"Number of conformations to generate (default: {configs.DEFAULT_NUMBER_CONFS})",
    )
    parser.add_argument(
        "-d",
        "--device",
        type=str,
        default=None,
        help="Device to use for predictions (default: None, auto-detected)",
    )
    parser.add_argument(
        "-s",
        "--steps",
        type=int,
        default=configs.DEFAULT_STEPS,
        help=f"Number of steps to run the DDPM model (default: {configs.DEFAULT_STEPS})",
    )
    parser.add_argument(
        "-m",
        "--method",
        type=str,
        default="mds",
        help=f"Method to use for generating 3D structures. Options are 'gd' or 'mds'. (default: {configs.DEFAULT_STRUCTURE_GEN})",
    )
    parser.add_argument(
        "-b",
        "--batch_size",
        type=int,
        default=configs.DEFAULT_BATCH_SIZE,
        help=f"Batch size to use for sampling (default: {configs.DEFAULT_BATCH_SIZE})",
    )
    parser.add_argument(
        "-o",
        "--output_directory",
        type=str,
        default=".",
        help="Directory to save output (default: '.')",
    )
    parser.add_argument(
        "--outname",
        type=str,
        default=None,
        help="If provided and a single sequence is provided, defines the prefix ahead of .pdb/.xtc/.npy extensions (default: None)",
    )
    parser.add_argument(
        "-r",
        "--return_structures",
        action="store_true",
        default=False,
        help="Return the 3D structures (default: False)",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Enable verbose output (default: False)",
    )
    parser.add_argument(
        "--num-cpus",
        dest="num_cpus",
        type=int,
        default=configs.DEFAULT_CPU_COUNT_MDS,
        help=f"Sets the max number of CPUs to use. Default: {configs.DEFAULT_CPU_COUNT_MDS}.",
    )
    parser.add_argument(
        "--num-mds-init",
        dest="num_mds_init",
        type=int,
        default=configs.DEFAULT_MDS_NUM_INIT,
        help=f"Sets the number of MDS jobs to be run in parallel. More may give better reconstruction but requires 1:1 with #CPUs to avoid performance penalty. Default: {configs.DEFAULT_MDS_NUM_INIT}.",
    )
    parser.add_argument(
        "--no-ddim",
        dest="ddim",
        default=True,
        action="store_false",
        help="Disable DDIM for sampling.",
    )
    parser.add_argument(
        "--disable_progress_bar",
        dest="progress_bar",
        action="store_false",
        default=True,
        help="Disable progress bar during generation (default: False)",
    )

    # will need to update this default...
    parser.add_argument(
        "--info",
        action="store_true",
        default=False,
        help="Print STARLING information only",
    )
    parser.add_argument(
        "--version",
        action="store_true",
        default=False,
        help="Print STARLING version only",
    )

    # Parse the command-line arguments
    args = parser.parse_args()

    # first. informational args overide anything else
    if args.info:
        print_info()
        sys.exit(0)
    elif args.version:
        print(__version__)
        sys.exit(0)
    elif args.user_input is None:
        print(parser.format_usage())  # Print the command signature
        print("ERROR: STARLING requires user_input or --version", file=sys.stderr)
        sys.exit(1)
    else:
        pass

    ### sanity checks
    # check if the output directory exists
    if not os.path.exists(args.output_directory):
        print(
            f"ERROR: Output directory {args.output_directory} does not exist.",
            file=sys.stderr,
        )
        sys.exit(1)

    if args.verbose:
        print_starling()

    # Call the generate function with parsed arguments
    generate(
        user_input=args.user_input,
        conformations=args.conformations,
        device=args.device,
        steps=args.steps,
        ddim=args.ddim,
        return_structures=args.return_structures,
        batch_size=args.batch_size,
        num_cpus_mds=args.num_cpus,
        num_mds_init=args.num_mds_init,
        output_directory=args.output_directory,
        output_name=args.outname,
        return_data=False,
        verbose=args.verbose,
        show_progress_bar=args.progress_bar,
    )


#
# BENCHMARKING CODE BELOW
#

# default sequence is alpha-synuclein
ALPHA_SYN = "MDVFMKGLSKAKEGVVAAAEKTKQGVAEAAGKTKEGVLYVGSKTKEGVVHGVATVAEKTKEQVTNVGGAVVTGVTAVAQKTVEGAGSIAAATGFVKKDQLGKNEEGAPQEGILEDMPVDPDNEAYEMPSEEGYQDYEPEA"


def measure_runtime(
    user_input, steps=25, batch_size=100, device=None, cooltime=600, single_run=0
):
    """
    Measure runtime of Starling conformational generation for a given sequence.

    Parameters
    ----------
    user_input : str
        Protein sequence to process.

    steps : int
        Number of steps to run the DDPM model.

    batch_size : int
        Batch size for processing.

    device : str
        Device to run on (e.g., 'cpu', 'cuda', 'mps').

    cooltime : int
        Time to cool down the hardware after each run (default: 600 seconds).

    single_run : int
        If set means we just do a single run with the specified number of conformations,
        otherwise we will do 10 predictions linearly spaced beteen 10 and 1000 confomers.

    """

    if single_run != 0:
        conformations_list = [int(single_run)]
    else:
        conformations_list = np.linspace(10, 1000, num=10, dtype=int)

    if device is None:
        device_name = str(check_device(device))
    else:
        device_name = device

    # initialize matrices to store runtime and radius of gyration
    runtime_matrix = np.zeros((len(conformations_list), 2))
    rg_matrix = np.zeros((len(conformations_list), 2))

    # iterate over conformations and steps
    for i, conf in enumerate(conformations_list):
        start_time = time.perf_counter()
        data = generate(
            user_input,
            conformations=conf,
            device=device,
            batch_size=batch_size,
            steps=steps,
            verbose=True,
            return_structures=False,
            return_single_ensemble=True,
        )
        end_time = time.perf_counter()

        runtime_matrix[i,0] = conf
        runtime_matrix[i,1] = end_time - start_time

        rg_matrix[i, 0] = conf
        rg_matrix[i, 1] = data.radius_of_gyration(return_mean=True)

        np.savetxt(
            f"rg_matrix_{steps}_steps_{device_name}_{batch_size}_batchsize.csv",
            rg_matrix,
            delimiter=", ",
        )
        np.savetxt(
            f"runtime_matrix_{steps}_steps_{device_name}_{batch_size}_batchsize.csv",
            runtime_matrix,
            delimiter=", ",
        )

        # have a snoozle to cool down the hardware (could use -80 alternatively? TBD).        
        if i != len(conformations_list) - 1:
            time.sleep(cooltime)


def starling_benchmark():
    parser = argparse.ArgumentParser(
        description="Run Starling conformational generation with specified parameters."
    )

    parser.add_argument(
        "--device",
        type=str,
        default=None,
        help="Device to run on (e.g., 'cpu', 'cuda', 'mps')",
    )
    parser.add_argument(
        "--batch-size",
        type=int,
        default=100,
        help="Batch size for processing (default: 100)",
    )
    parser.add_argument(
        "--steps",
        type=int,
        default=25,
        help="Number of steps to run the DDPM model (default: 25)",
    )
    parser.add_argument(
        "--sequence",
        type=str,
        default=ALPHA_SYN,
        help="Protein sequence to process. Default is alpha-synuclein.",
    )
    parser.add_argument(
        "--cooltime",
        type=int,
        default=20,
        help="Time to cool down the hardware after each run (default: 20 seconds).",
    )
    parser.add_argument(
        "--single-run",
        type=int,
        default=0,
        help="By default, starling-benchmark runs a series of 10 predictions for 10-1000 confomers. If you want to run a single prediction, specify the number of conformations here.",
    )

    args = parser.parse_args()

    measure_runtime(
        args.sequence,
        steps=args.steps,
        batch_size=args.batch_size,
        device=args.device,
        cooltime=args.cooltime,
        single_run=args.single_run,
    )



import torch
from einops import rearrange
from torch import nn

from starling.data.positional_encodings import (
    PositionalEncoding1D,
    PositionalEncoding2D,
)
from starling.models.attention import CrossAttention, SelfAttention


class GeGLU(nn.Module):
    def __init__(self, d_in: int, d_out: int):
        """
        Activation function that combines the concept of gating with the GELU activation function.
        The gating mechanism is used to control the flow of information through the network. The GELU
        activation function is used to introduce non-linearity in the network. The GeGLU activation
        function is often seen in the feed forward layer of transformers.

        The GeGLU activation function
        is defined as follows: x * GELU(gate), where x is the input to the activation function and
        gate is the output of a linear layer.

        Parameters
        ----------
        d_in : int
            The input dimension of the data. Used to initialize the linear layer.
        d_out : int
            The output dimension of the data. Used to initialize the linear layer.
        """
        super().__init__()

        self.proj = nn.Linear(d_in, d_out * 2)
        self.gelu = nn.GELU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x, gate = self.proj(x).chunk(2, dim=-1)
        return x * self.gelu(gate)


class FeedForward(nn.Module):
    def __init__(self, embed_dim: int):
        """
        Feed forward layer in the transformer architecture. The feed forward layer consists of
        two linear layers with a GELU activation function in between. The linear layers first
        expand the number of dimensions by a factor of 4 and then reduce the number of dimensions
        back to the original number of dimensions. The GELU activation function is used to introduce
        non-linearity in the network.

        Parameters
        ----------
        embed_dim : int
            The input dimension of the data. Used to initialize the linear layers.
        """
        super().__init__()

        self.net = nn.Sequential(
            GeGLU(embed_dim, embed_dim * 4),
            nn.Linear(embed_dim * 4, embed_dim),
        )

        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if x.dim() == 4:
            x = rearrange(x, "b c h w -> b h w c")

        x = self.net(self.norm(x))

        if x.dim() == 4:
            x = rearrange(x, "b h w c -> b c h w")
        return x


class TransformerEncoder(nn.Module):
    def __init__(self, embed_dim: int, num_heads: int, custom_attention: bool):
        """
        Transformer encoder layer. The transformer encoder layer consists of a self attention layer
        and a feed forward layer. The self attention layer is used to capture the relationships
        between different elements in the input data. The feed forward layer is used to introduce
        non-linearity in the network.

        Parameters
        ----------
        embed_dim : int
            The input dimension of the data. Used to initialize the self attention and feed forward layers.
        num_heads : int
            The number of heads in the multi-head attention layer. Used to initialize the self attention layer.
        """
        super().__init__()

        self.self_attention = SelfAttention(
            embed_dim, num_heads, custom=custom_attention
        )
        self.feed_forward = FeedForward(embed_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Prenorm is happening within the self attention layer
        x = x + self.self_attention(x)

        # Prenorm is happening within the feed forward layer
        x = x + self.feed_forward(x)
        return x


class TransformerDecoder(nn.Module):
    def __init__(self, embed_dim: int, num_heads: int, context_dim, custom_attention):
        """
        Transformer decoder layer. The transformer decoder layer consists of a self attention layer,
        cross attention layer and a feed forward layer. The self attention layer is used to capture the
        relationships between different elements in the input data. The cross attention layer is used to
        capture the relationships between the input data and the context data (usually the transformer
        encoder output). The feed forward layer is used to introduce non-linearity in the network.

        Parameters
        ----------
        embed_dim : int
            The input dimension of the data. Used to initialize the self attention, cross attention and feed forward layers.
        num_heads : int
            The number of heads in the multi-head attention layer. Used to initialize the self attention and cross attention layers.
        context_dim : _type_
            The dimension of the context data. Used to initialize the cross attention layer.
        """
        super().__init__()

        self.self_attention = SelfAttention(
            embed_dim, num_heads, custom=custom_attention
        )
        self.cross_attention = CrossAttention(
            embed_dim, num_heads, context_dim, custom=custom_attention
        )
        self.feed_forward = FeedForward(embed_dim)

    def forward(self, x: torch.Tensor, context=None) -> torch.Tensor:
        # Prenorm is happening within the self attention layer
        x = x + self.self_attention(x)

        # Prenorm is happening within the cross attention layer
        x = x + self.cross_attention(x, context)

        # Prenorm is happening within the feed forward layer
        x = x + self.feed_forward(x)
        return x


class SpatialTransformer(nn.Module):
    def __init__(
        self, embed_dim: int, num_heads: int, context_dim: int, custom_attention: bool
    ):
        """
        Spatial transformer network. The spatial transformer network consists of a transformer encoder
        and a transformer decoder. The transformer encoder is used to process the features of the
        context data (in our case protein sequences). The transformer decoder is used to capture the relationships
        between the input data and the context data. The spatial transformer network is used to generate
        the latent space representation of the input data.

        Parameters
        ----------
        embed_dim : int
            The input dimension of the data. Used to initialize the transformer encoder and decoder.
        num_heads : int
            The number of heads in the multi-head attention layer. Used to initialize the transformer encoder and decoder.
        context_dim : int
            The dimension of the context data. Used to initialize the transformer encoder and decoder.
        """
        super().__init__()

        # Add positional encodings to the context (protein sequence data)
        self.context_positional_encodings = PositionalEncoding1D(384, context_dim)
        self.context_encoder = TransformerEncoder(
            context_dim, num_heads, custom_attention=custom_attention
        )

        # Add positional encodings to the latent space representation of images (e.i. distance maps)
        self.image_positional_encodings = PositionalEncoding2D(embed_dim)
        self.group_norm = nn.GroupNorm(num_groups=32, num_channels=embed_dim)
        self.conv_in = nn.Conv2d(embed_dim, embed_dim, kernel_size=1)
        self.transformer_block = TransformerDecoder(
            embed_dim, num_heads, context_dim, custom_attention=custom_attention
        )
        self.conv_out = nn.Conv2d(embed_dim, embed_dim, kernel_size=1)

    def forward(self, x: torch.Tensor, context=None) -> torch.Tensor:
        # Save the input for the residual connection
        x_in = x

        # Add positional encodings to the context and process the context features
        context = self.context_positional_encodings(context)
        context = self.context_encoder(context)

        # Add positional encodings to the latent space representation of images
        x = self.image_positional_encodings(x)
        x = self.group_norm(x)
        x = self.conv_in(x)

        # Transformer block to capture the relationships between the input data and the context data
        x = self.transformer_block(x, context)

        x = self.conv_out(x)

        # Residual connection
        return x + x_in



import os

import numpy as np
import protfasta
from tqdm import tqdm

#
from starling import configs, utilities
from starling.inference import generation


def handle_input(
    user_input, invalid_sequence_action="convert", output_name=None, seq_index_start=1
):
    """
    Dynamically handle the input from the user.
    This returns a dictionary with either the names from
    the user's input file or the users input dictionary of
    sequences or will create a dictionary with the sequences
    numbered in the order they were passed in with seq_index_start
    as the starting index.

    Parameters
    -----------
    user_input: str, list, dict
        This can be one of a few different options:
            str: A .fasta file
            str: A seq.in file formatted as a .tsv with name\tseq
            str: A .tsv file formatted as name\tseq. Same as seq.in
                except a different file extension. Borna used a seq.in
                in his tutorial, so I'm rolling with it.
            str: A sequence as a string
            list: A list of sequences
            dict: A dict of sequences (name: seq)

    invalid_sequence_action: str
        This can be one of 3 options:
            fail - invalid sequence cause parsing to fail and throw an exception
            remove - invalid sequences are removed
            convert - invalid sequences are converted
            Default is 'convert'
            Only these 3 options are allowed because STARLING cannot handle
            non-canonical residues, so we don't want to use the protfasta.read_fasta()
            options that allow this to happen.

    output_name : str
        If provided and if a single amino acid sequence is passed in, this will be the key
        in the output dictionary. If None, the key will be 'sequence_<index>'. If a dictionary
        or list or path to a FASTA file is passed, this is ignored. Default is None.

    seq_index_start: int
        If we need to number sequences in the output dictionary, this is the starting index.
        This is only needed if a sequence as a string is passed in or if a list of sequences
        is passed in.

    Returns
    --------
    dict: A dictionary of sequences (name: seq)
    """

    # Helper function to validate and clean sequences.
    # This will raise an Exception if the sequence contains non-valid amino acids.
    # and makes sure everything is uppercase.
    def clean_sequence(sequence):
        sequence = sequence.upper()
        valid_residues = set("ACDEFGHIKLMNPQRSTVWY")  # Standard amino acids
        cleaned = "".join([res for res in sequence if res in valid_residues])
        # check lengths
        if len(cleaned) != len(sequence):
            raise ValueError(f"Invalid amino acid detected in sequence: {sequence}")
        # return the cleaned sequence.
        return cleaned

    ### Check and handle different input types

    # If input is a string...
    if isinstance(user_input, str):
        if user_input.endswith((".fasta", ".FASTA", ".tsv", ".in")):
            # make sure user has a valid path.
            if not os.path.exists(user_input):
                raise FileNotFoundError(f"File {user_input} not found.")
            # if a .fasta, use protfasta.
            if user_input.endswith((".fasta", ".FASTA")):
                # this will throw an error if we have duplicate sequence names, so
                # don't need to worry about that here.
                sequence_dict = protfasta.read_fasta(
                    user_input, invalid_sequence_action=invalid_sequence_action
                )
            elif user_input.endswith((".tsv", ".in")):
                # this doesn't have a check for duplicate sequences names,
                # so we should do that here. This should be the only instance
                # where that might cause an issue. Current behavior is to force duplicate names.
                sequence_dict = {}
                with open(user_input, "r") as f:
                    for line in f:
                        name, seq = line.strip().split("\t")
                        if name not in sequence_dict:
                            sequence_dict[name] = clean_sequence(seq)
                        else:
                            raise ValueError(
                                f"Duplicate sequence name detected: {name}.\nPlease ensure sequences have unique names in the input file."
                            )
            return sequence_dict
        else:
            # otherwise only string input allowed is a sequence as a string. Automatically create
            # the name if output_name is None.
            if output_name is None:
                return {f"sequence_{seq_index_start}": clean_sequence(user_input)}

            # if output_name is not None, use that as the key in the dictionary.
            else:
                try:
                    output_name = str(output_name)
                except Exception as e:
                    raise ValueError(
                        "output_name must be a string our castable to a string."
                    )
                return {output_name: clean_sequence(user_input)}

    # if input is a list
    elif isinstance(user_input, list):
        # if a list, make sure all sequences are valid.
        sequence_dict = {}
        for i, seq in enumerate(user_input):
            sequence_dict[f"sequence_{i + seq_index_start}"] = clean_sequence(seq)
        return sequence_dict
    elif isinstance(user_input, dict):
        # if a dict, make sure all sequences are valid.
        sequence_dict = {}
        for name, seq in user_input.items():
            sequence_dict[name] = clean_sequence(seq)
        return sequence_dict
    else:
        raise ValueError(
            f"Invalid input type: {type(user_input)}. Must be str, list, or dict."
        )


def check_positive_int(val):
    """
    Function to check if a value is a positive integer.

    Parameters
    ---------------
    val : int
        The value to check.

    Returns
    ---------------
    bool: True if val is a positive integer, False otherwise.
    """
    if isinstance(val, int) or np.issubdtype(type(val), np.integer):
        if val > 0:
            return True
    return False


def generate(
    user_input,
    conformations=configs.DEFAULT_NUMBER_CONFS,
    device=None,
    steps=configs.DEFAULT_STEPS,
    ddim=True,
    return_structures=False,
    batch_size=configs.DEFAULT_BATCH_SIZE,
    num_cpus_mds=configs.DEFAULT_CPU_COUNT_MDS,
    num_mds_init=configs.DEFAULT_MDS_NUM_INIT,
    output_directory=None,
    output_name=None,
    return_data=True,
    verbose=False,
    show_progress_bar=True,
    show_per_step_progress_bar=True,
    pdb_trajectory=False,
    return_single_ensemble=False,
):
    """
    Main function for generating the distance maps using STARLING. Allows
    you to pass a single sequence, a list of sequences, a dictionary, or
    a path to a .fasta file, a .tsv file, or a seq.in file, and from that
    generate distance maps and 3D conformational ensembles using the
    STARLING model. This function is the main user-facing STARLING function.

    Note: if you want to change the location of the networks,
    you need to change them in the configs.py file. Those paths get
    read in by the ModelManager class and are not passed in as arguments
    to this function. This lets us avoid iteratively loading the network
    when running the generate function multiple times in a single session.

    Parameters
    ---------------
    user_input : str, list, dict
        This can be one of a few different options:
            str: A path to a .fasta file as a str.
            str: A path to a seq.in file formatted as a .tsv with name\tseq
            str: A path to a .tsv file formatted as name\tseq. Same as
                seq.in except a different file extension. Borna used a seq.in
                in his tutorial, so I'm rolling with it.
            str: A sequence as a string
            list: A list of sequences
            dict: A dict of sequences (name: seq)

    conformations : int
        The number of conformations to generate. Default is 200.

    device : str
        The device to use for predictions. Default is None. If None, the
        default device will be used. Default device is 'gpu'.
        This is MPS for apple silicon and CUDA for all other devices.
        If MPS and CUDA are not available, automatically falls back to CPU.

    steps : int
        The number of steps to run the DDPM model. Default is 10.

    ddim : bool
        Whether to use DDIM for sampling. Default is True.

    return_structures : bool
        Whether to return the 3D structure. Default is False.

    batch_size : int
        The batch size to use for sampling. 100 uses ~20 GB
        of memory. Default is 100.

    num_cpus_mds : int
        The number of CPUs to use for MDS. Default is 4

    num_mds_init : int
        Number of independent MDS jobs to execute. Default is
        4. Note if goes up in principle more shots of finding
        a good solution but there is a performance hit unless
        num_cpus_mds == num_mds_init.

    output_directory : str
        The path to save the output.

        If set to None, no output will be saved to disk.

        If not None, will save the output to the specified path.
        This includes the distance maps and, if
        return_structures=True, the 3D structures.

        The distance maps are saved as .npy files with the names
        <sequence_name>_STARLING_DM.npy and the structures are
        saved with the file names <sequence_name>_STARLING.xtc
        and <sequence_name>_STARLING.pdb.

        <sequence_name> here will depend on the input provided
        to generate. If the input is a dictionary, then the keys
        will be used as the sequence names. If the input is a
        list or a single sequence, the sequence will be saved
        as 'sequence_<index>'. If a path to a FASTA file is passed
        in, the headers from the FASTA file will be used. Note also
        if a single sequence is passed the sequence_<index> format
        can be overridden by setting the output_name parameter.

        Default is None.

    output_name : str
        If provided and if a single amino acid sequence is passed in,
        this will be the key in the output dictionary. If None, the
        key will be 'sequence_<index>'. If a dictionary or list or path
        to a FASTA file is passed, this is ignored. Default is None.

    return_data : bool
        If True, will return a dictionary of Ensemble objects which will
        include structural ensembles if return_structures=True. If False,
        will return None (so you need to set the output_dictionary, or the
        analysis will be lost!). Default is True.

    verbose : bool
        Whether to print verbose output. Default is False.

    show_progress_bar : bool
        Whether to show a progress bar. Default is True.

    show_per_step_progress_bar : bool, optional
        Whether to show progress bar per step.
        Default is True
    pdb_trajectory : bool
        Whether to save the trajectory as a pdb file.
        Default is False.

    return_single_ensemble : bool
        If True, will return a single starling.structure.ensemble.Ensemble
        object instead of a dictionary of ensemble objects IF and only if
        there is one sequence passed. If this option is passed and multiple
        sequences are passed this will throw an ValueError. Default False.

    Returns
    ---------------
    dict, None, or Ensemble
        The function returns a dictionary of Ensemble objects,
        a single Ensemble object, or a None depending on the
        requested return information.

        The default behavior is to return a dict of Ensemble
        objects, which happens if return_data=True and
        and return_single_ensemble=False.

        If return_data=False then None is returned.

        If return_single_ensemble=True then a single Ensemble
        object is returned.

        if return_data=False and return_single_ensemble=True
        then a ValueError exception is raised.

    """
    # check user input, return a sequence dict.
    _sequence_dict = handle_input(user_input, output_name=output_name)

    # we do this specific sanity check EARLY so we don't silently fix what would
    # otherwise be a faulty input
    if return_single_ensemble and len(_sequence_dict) > 1:
        raise ValueError(
            f"Error: requested single ensemble yet provided input of {len(_sequence_dict)} sequences."
        )

    # filter out sequences that are too long (rather than erroring out)
    sequence_dict = {}
    removed_counter = 0
    for k in _sequence_dict:
        if len(_sequence_dict[k]) > configs.MAX_SEQUENCE_LENGTH:
            print(
                f"Warning: Sequence {k} is too long; maximum sequence in STARLING is {configs.MAX_SEQUENCE_LENGTH} residues, {k} is {len(_sequence_dict[k])}. Skipping..."
            )
            removed_counter = removed_counter + 1
        else:
            sequence_dict[k] = _sequence_dict[k]

    if verbose:
        # if we removed one sequence for being too long...
        if removed_counter == 1:
            bonus_message = f". Removed {removed_counter} sequence for being too long"

        # if we removed more than one sequence for being too long....
        elif removed_counter > 1:
            bonus_message = f". Removed {removed_counter} sequences for being too long"

        # if we removed no sequences!
        else:
            bonus_message = ""

        if len(sequence_dict) == 1:
            print(f"[STATUS]: Generating distance maps for 1 sequence{bonus_message}.")
        else:
            print(
                f"[STATUS]: Generating distance maps for {len(sequence_dict)} sequences{bonus_message}."
            )

    # check various other things so we fail early. Don't
    # want to go about the entire process and then have it fail at the end.
    # check conformations
    if not check_positive_int(conformations):
        raise ValueError("Error: Conformations must be an integer greater than 0.")

    # check steps
    if not check_positive_int(steps):
        raise ValueError("Error: Steps must be an integer greater than 0.")

    # check batch size
    if not check_positive_int(batch_size):
        raise ValueError("Error: batch_size must be an integer greater than 0.")

    # check number of cpus
    if not check_positive_int(num_cpus_mds):
        raise ValueError("Error: num_cpus_mds must be an integer greater than 0.")

    # check number of independent runs of MDS
    if not check_positive_int(num_mds_init):
        raise ValueError("Error: num_mds_init must be an integer greater than 0.")

    # make sure batch_size is not smaller than conformations.
    # if it is, make batch_size = conformations.
    if batch_size > conformations:
        batch_size = conformations

    # check output_directory is a directory that exists.
    if output_directory is not None:
        if not os.path.exists(output_directory):
            raise FileNotFoundError(
                f"Error: Directory {output_directory} does not exist."
            )

    # check ddim is a bool
    if not isinstance(ddim, bool):
        raise ValueError("Error: DDIM must True or False.")

    # check return_structures is a bool
    if not isinstance(return_structures, bool):
        raise ValueError("Error: return_structures must be True or False.")

    # check verbose is a bool
    if not isinstance(verbose, bool):
        raise ValueError("Error: verbose must be True or False.")

    # check show_progress_bar
    if not isinstance(show_progress_bar, bool):
        raise ValueError("Error: show_progress_bar must be True or False.")

    # check show_per_step_progress_bar
    if not isinstance(show_per_step_progress_bar, bool):
        raise ValueError("Error: show_per_step_progress_bar must be True or False.")

    # we do this specific sanity check to make the logic later in this function easier
    if return_single_ensemble and return_data is False:
        raise ValueError(
            "Error: requested single ensemble yet also did not request data to be returned."
        )

    if return_data is False and output_directory is None:
        raise ValueError(
            "Error: both no return data (return_data=False) and also did not specifiy an output_directory; this means no output will be returned/saved anywhere, which is probably not desired!"
        )

    # check device, get back a torch.device (not a str!)
    device = utilities.check_device(device)

    # run the actual inference and return the results
    ensemble_return = generation.generate_backend(
        sequence_dict,
        conformations,
        device,
        steps,
        ddim,
        return_structures,
        batch_size,
        num_cpus_mds,
        num_mds_init,
        output_directory,
        return_data,
        verbose,
        show_progress_bar,
        show_per_step_progress_bar,
        pdb_trajectory,
    )

    # if this is true we KNOW there is only one Ensemble in the return dict because
    # we previously checked for this.
    if return_single_ensemble:
        return list(ensemble_return.values())[0]
    else:
        return ensemble_return
